# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("cryptnet.dll")
prototypes = \
    {
        #
        'CryptRetrieveObjectByUrlA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "pszCredentialsOid": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "pvCredentials": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="CRYPT_CREDENTIALS", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "pLastSyncTime": SimTypePointer(SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), offset=0), "dwMaxUrlRetrievalByteCount": SimTypeInt(signed=False, label="UInt32"), "pPreFetchInfo": SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "dwObjectType": SimTypeInt(signed=False, label="UInt32"), "dwError": SimTypeInt(signed=False, label="UInt32"), "dwReserved": SimTypeInt(signed=False, label="UInt32"), "ThisUpdateTime": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "NextUpdateTime": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "PublishTime": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None)}, name="CRYPTNET_URL_CACHE_PRE_FETCH_INFO", pack=False, align=None), offset=0), "pFlushInfo": SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "dwExemptSeconds": SimTypeInt(signed=False, label="UInt32"), "ExpireTime": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None)}, name="CRYPTNET_URL_CACHE_FLUSH_INFO", pack=False, align=None), offset=0), "ppResponseInfo": SimTypePointer(SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wResponseType": SimTypeShort(signed=False, label="UInt16"), "wResponseFlags": SimTypeShort(signed=False, label="UInt16"), "LastModifiedTime": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "dwMaxAge": SimTypeInt(signed=False, label="UInt32"), "pwszETag": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwProxyId": SimTypeInt(signed=False, label="UInt32")}, name="CRYPTNET_URL_CACHE_RESPONSE_INFO", pack=False, align=None), offset=0), offset=0), "pwszCacheFileNamePrefix": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pftCacheResync": SimTypePointer(SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), offset=0), "fProxyCacheRetrieval": SimTypeInt(signed=True, label="Int32"), "dwHttpStatusCode": SimTypeInt(signed=False, label="UInt32"), "ppwszErrorResponseHeaders": SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0), "ppErrorContentBlob": SimTypePointer(SimTypePointer(SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPTOAPI_BLOB", pack=False, align=None), offset=0), offset=0)}, name="CRYPT_RETRIEVE_AUX_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pszUrl", "pszObjectOid", "dwRetrievalFlags", "dwTimeout", "ppvObject", "hAsyncRetrieve", "pCredentials", "pvVerify", "pAuxInfo"]),
        #
        'CryptRetrieveObjectByUrlW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "pszCredentialsOid": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "pvCredentials": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="CRYPT_CREDENTIALS", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "pLastSyncTime": SimTypePointer(SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), offset=0), "dwMaxUrlRetrievalByteCount": SimTypeInt(signed=False, label="UInt32"), "pPreFetchInfo": SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "dwObjectType": SimTypeInt(signed=False, label="UInt32"), "dwError": SimTypeInt(signed=False, label="UInt32"), "dwReserved": SimTypeInt(signed=False, label="UInt32"), "ThisUpdateTime": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "NextUpdateTime": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "PublishTime": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None)}, name="CRYPTNET_URL_CACHE_PRE_FETCH_INFO", pack=False, align=None), offset=0), "pFlushInfo": SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "dwExemptSeconds": SimTypeInt(signed=False, label="UInt32"), "ExpireTime": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None)}, name="CRYPTNET_URL_CACHE_FLUSH_INFO", pack=False, align=None), offset=0), "ppResponseInfo": SimTypePointer(SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "wResponseType": SimTypeShort(signed=False, label="UInt16"), "wResponseFlags": SimTypeShort(signed=False, label="UInt16"), "LastModifiedTime": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "dwMaxAge": SimTypeInt(signed=False, label="UInt32"), "pwszETag": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwProxyId": SimTypeInt(signed=False, label="UInt32")}, name="CRYPTNET_URL_CACHE_RESPONSE_INFO", pack=False, align=None), offset=0), offset=0), "pwszCacheFileNamePrefix": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pftCacheResync": SimTypePointer(SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), offset=0), "fProxyCacheRetrieval": SimTypeInt(signed=True, label="Int32"), "dwHttpStatusCode": SimTypeInt(signed=False, label="UInt32"), "ppwszErrorResponseHeaders": SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0), "ppErrorContentBlob": SimTypePointer(SimTypePointer(SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="CRYPTOAPI_BLOB", pack=False, align=None), offset=0), offset=0)}, name="CRYPT_RETRIEVE_AUX_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pszUrl", "pszObjectOid", "dwRetrievalFlags", "dwTimeout", "ppvObject", "hAsyncRetrieve", "pCredentials", "pvVerify", "pAuxInfo"]),
        #
        'CryptInstallCancelRetrieval': SimTypeFunction([SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dwFlags", "pvArg"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pfnCancel", "pvArg", "dwFlags", "pvReserved"]),
        #
        'CryptUninstallCancelRetrieval': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dwFlags", "pvReserved"]),
        #
        'CryptGetObjectUrl': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="CRYPT_GET_URL_FLAGS"), SimTypePointer(SimStruct({"cUrl": SimTypeInt(signed=False, label="UInt32"), "rgwszUrl": SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0)}, name="CRYPT_URL_ARRAY", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=False, label="UInt32"), "dwSyncDeltaTime": SimTypeInt(signed=False, label="UInt32"), "cGroup": SimTypeInt(signed=False, label="UInt32"), "rgcGroupEntry": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="CRYPT_URL_INFO", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pszUrlOid", "pvPara", "dwFlags", "pUrlArray", "pcbUrlArray", "pUrlInfo", "pcbUrlInfo", "pvReserved"]),
    }

lib.set_prototypes(prototypes)



#!/usr/bin/env python3
__package__ = __package__ or "tests.simos.windows"  # pylint:disable=redefined-builtin

import os
import struct
import unittest

import angr
import angr.simos.windows

from ...common import bin_location


test_location = os.path.join(bin_location, "tests")


def compare_none(state, test_value):
    test_value = test_value.concrete
    correct_value = angr.simos.windows.VS_SECURITY_COOKIES[state.arch.name].default
    assert test_value == correct_value


def compare_random(state, test_value):
    test_value = test_value.concrete
    incorrect_value = angr.simos.windows.VS_SECURITY_COOKIES[state.arch.name].default
    assert test_value != incorrect_value


def compare_static(state, test_value):
    test_value = test_value.concrete
    correct_value = struct.unpack(">I", b"cook")[0]
    assert test_value == correct_value


def compare_symbolic(state, test_value):
    assert test_value.resolved.symbolic


def check_value(project, init_type, comparison):
    main_object = project.loader.main_object
    state = project.factory.blank_state(security_cookie_init=init_type)
    value = getattr(state.mem[main_object.load_config["SecurityCookie"]], f"uint{state.arch.bits}_t")
    comparison(state, value)


class TestWindowsStackCookie(unittest.TestCase):
    def test_security_cookie_init(self):
        project = angr.Project(os.path.join(test_location, "i386", "test_arrays.exe"), auto_load_libs=False)
        check_value(project, angr.simos.windows.SecurityCookieInit.NONE, compare_none)
        check_value(project, angr.simos.windows.SecurityCookieInit.RANDOM, compare_random)
        check_value(project, angr.simos.windows.SecurityCookieInit.STATIC, compare_static)
        check_value(project, angr.simos.windows.SecurityCookieInit.SYMBOLIC, compare_symbolic)

        self.assertRaises(TypeError, project.factory.blank_state, security_cookie_init=1)


if __name__ == "__main__":
    unittest.main()



from . import SimConcretizationStrategy


class SimConcretizationStrategyRange(SimConcretizationStrategy):
    """
    Concretization strategy that resolves addresses to a range.
    """

    def __init__(self, limit, **kwargs):  # pylint:disable=redefined-builtin
        super().__init__(**kwargs)
        self._limit = limit

    def _concretize(self, memory, addr, **kwargs):
        mn, mx = self._range(memory, addr, **kwargs)
        if mx - mn <= self._limit:
            return self._eval(memory, addr, self._limit, **kwargs)
        return None



#!/usr/bin/env python3
import os
import unittest

from angr.analyses.cfg_slice_to_sink import CFGSliceToSink
from angr.project import Project


class TestCFGSliceToSink(unittest.TestCase):
    def test_get_transitions_from_slice(self):
        transitions = {1: [2, 3]}
        my_slice = CFGSliceToSink(None, transitions)

        self.assertDictEqual(my_slice.transitions, transitions)

    def test_get_entrypoints_from_slice(self):
        transitions = {0: [2], 1: [2, 3], 2: [4, 5]}
        my_slice = CFGSliceToSink(None, transitions)

        self.assertListEqual(my_slice.entrypoints, [0, 1])

    def test_add_transitions_updates_the_slice(self):
        my_slice = CFGSliceToSink(None, {1: [2, 3]})
        transitions_to_add = {1: [4], 2: [4]}

        result = my_slice.add_transitions(transitions_to_add)

        expected_result = {
            1: [2, 3, 4],
            2: [4],
        }

        self.assertDictEqual(result, expected_result)

    def test_nodes(self):
        my_slice = CFGSliceToSink(
            None,
            {
                1: [2, 3],
                2: [3],
            },
        )

        expected_result = [1, 2, 3]
        result = my_slice.nodes

        self.assertListEqual(result, expected_result)

    def test_transitions_as_tuples(self):
        my_slice = CFGSliceToSink(None, {1: [2, 3], 2: [3]})

        expected_result = [(1, 2), (1, 3), (2, 3)]
        result = my_slice.transitions_as_tuples

        self.assertListEqual(result, expected_result)

    def disable_emptyness(self):
        # disabled since binaries-private is not checked out for angr CI
        binaries_path = os.path.join(
            os.path.dirname(__file__), "..", "..", "..", "..", "binaries-private", "operation-mango"
        )
        binary_path = os.path.join(binaries_path, "air-live-bu-2015", "cgi_test.cgi")
        project = Project(binary_path, auto_load_libs=False)
        cfg = project.analyses.CFGFast()
        printf = cfg.kb.functions.function(name="printf", plt=False)
        printf_node = cfg.model.get_all_nodes(printf.addr)[0]

        printf_predecessor = printf_node.predecessors[0]

        empty_slice = CFGSliceToSink(printf, {})
        non_empty_slice = CFGSliceToSink(printf, {printf_predecessor.addr: [printf.addr]})

        self.assertEqual(empty_slice.is_empty(), True)
        self.assertEqual(non_empty_slice.is_empty(), False)

    def test_path_between_returns_True_only_if_there_exists_at_least_a_path_between_two_nodes_in_the_slice(self):
        my_slice = CFGSliceToSink(None, {1: [2, 3], 2: [4]})

        self.assertTrue(my_slice.path_between(1, 2))
        self.assertTrue(my_slice.path_between(1, 3))
        self.assertTrue(my_slice.path_between(2, 4))
        self.assertTrue(my_slice.path_between(1, 4))

        self.assertFalse(my_slice.path_between(3, 4))

    def test_path_between_deals_with_loops(self):
        my_slice = CFGSliceToSink(None, {1: [2, 3], 2: [1]})

        self.assertFalse(my_slice.path_between(1, 4))


if __name__ == "__main__":
    unittest.main()



import claripy

from . import MemoryMixin


class ConditionalMixin(MemoryMixin):
    def load(self, addr, condition=None, fallback=None, **kwargs):
        res = super().load(addr, condition=condition, **kwargs)
        if condition is not None and fallback is not None:
            res = claripy.If(condition, res, fallback)
        return res

    def store(self, addr, data, size=None, condition=None, **kwargs):
        condition = self.state._adjust_condition(condition)

        if condition is None or self.state.solver.is_true(condition):
            super().store(addr, data, size=size, **kwargs)
            return
        if self.state.solver.is_false(condition):
            return

        default_data = super().load(addr, size=len(data) // self.state.arch.byte_width, **kwargs)
        conditional_data = claripy.If(condition, data, default_data)
        super().store(addr, conditional_data, size=size, **kwargs)



import angr
from collections import namedtuple

import logging

l = logging.getLogger(name=__name__)

Dirent = namedtuple("dirent", ("d_ino", "d_off", "d_reclen", "d_type", "d_name"))


class readdir(angr.SimProcedure):
    struct = None
    condition = None

    def run(self, dirp):  # pylint: disable=arguments-differ
        # TODO: make sure argument is actually a dir struct
        if self.state.arch.name != "AMD64":
            l.error("readdir SimProcedure is only implemented for AMD64")
            return 0

        self._build_amd64()
        self.instrument()
        malloc = angr.SIM_PROCEDURES["libc"]["malloc"]
        pointer = self.inline_call(malloc, 19 + 256).ret_expr
        self._store_amd64(pointer)
        return self.state.solver.If(self.condition, pointer, 0)

    def instrument(self):
        """
        Override this function to instrument the SimProcedure.

        The two useful variables you can override are self.struct, a named tuple of all the struct
        fields, and self.condition, the condition for whether the function succeeds.
        """
        pass

    def _build_amd64(self):
        self.struct = Dirent(
            self.state.solver.BVV(0, 64),  # d_ino
            self.state.solver.BVV(0, 64),  # d_off
            self.state.solver.BVS("d_reclen", 16, key=("api", "readdir", "d_reclen")),  # d_reclen
            self.state.solver.BVS("d_type", 8, key=("api", "readdir", "d_type")),  # d_type
            self.state.solver.BVS("d_name", 255 * 8, key=("api", "readdir", "d_name")),
        )  # d_name
        self.condition = self.state.solver.BoolS("readdir_cond")  # TODO: variable key

    def _store_amd64(self, ptr):
        def stores(offset, val):
            return self.state.memory.store(ptr + offset, val, endness="Iend_BE")

        def storei(offset, val):
            return self.state.memory.store(ptr + offset, val, endness="Iend_LE")

        storei(0, self.struct.d_ino)
        storei(8, self.struct.d_off)
        storei(16, self.struct.d_reclen)
        storei(18, self.struct.d_type)
        stores(19, self.struct.d_name)
        stores(19 + 255, self.state.solver.BVV(0, 8))



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,no-self-use,line-too-long
__package__ = __package__ or "tests.procedures"  # pylint:disable=redefined-builtin

import os
import unittest

import angr

from ..common import bin_location


test_location = os.path.join(bin_location, "tests")
bina = os.path.join(test_location, "x86_64", "test_project_resolve_simproc")


# We voluntarily don't use SimProcedures for 'rand' and 'sleep' because we want
# to step into their lib code.
class TestProjectResolveSimProc(unittest.TestCase):
    def test_bina(self):
        # auto_load_libs can't be disabled as the testcase fails
        p = angr.Project(bina, exclude_sim_procedures_list=["rand", "sleep"], load_options={"auto_load_libs": True})

        # Make sure external functions are not replaced with a SimProcedure
        sleep_jmpslot = p.loader.main_object.jmprel["sleep"]
        rand_jmpslot = p.loader.main_object.jmprel["rand"]
        read_jmpslot = p.loader.main_object.jmprel["read"]

        sleep_addr = p.loader.memory.unpack_word(sleep_jmpslot.rebased_addr)
        rand_addr = p.loader.memory.unpack_word(rand_jmpslot.rebased_addr)
        read_addr = p.loader.memory.unpack_word(read_jmpslot.rebased_addr)

        libc_sleep_addr = p.loader.shared_objects["libc.so.6"].get_symbol("sleep").rebased_addr
        libc_rand_addr = p.loader.shared_objects["libc.so.6"].get_symbol("rand").rebased_addr

        assert sleep_addr == libc_sleep_addr
        assert rand_addr == libc_rand_addr
        assert p.is_hooked(read_addr)
        assert "read" in str(p._sim_procedures[read_addr])


if __name__ == "__main__":
    unittest.main()



import logging
from collections import defaultdict

from . import ExplorationTechnique


l = logging.getLogger(name=__name__)


class LocalLoopSeer(ExplorationTechnique):
    """
    LocalLoopSeer monitors exploration and maintains all loop-related data without relying on a control flow graph.
    """

    def __init__(self, bound=None, bound_reached=None, discard_stash="spinning"):
        """
        :param bound:                 Limit the number of iterations a loop may be executed.
        :param bound_reached:         If provided, should be a function that takes the LoopSeer and the succ_state.
                                      Will be called when loop execution reach the given bound.
                                      Default to moving states that exceed the loop limit to a discard stash.
        :param discard_stash:         Name of the stash containing states exceeding the loop limit.
        """

        super().__init__()
        self.bound = bound
        self.bound_reached = bound_reached
        self.discard_stash = discard_stash
        self.block_counters = defaultdict(int)
        self.cut_succs = []

    def setup(self, simgr):
        pass

    def filter(self, simgr, state, **kwargs):
        if state in self.cut_succs:
            self.cut_succs.remove(state)
            return self.discard_stash
        else:
            return simgr.filter(state, **kwargs)

    def successors(self, simgr, state, **kwargs):
        succs = simgr.successors(state, **kwargs)

        for succ_state in succs.successors:
            # Processing a currently running loop

            if succ_state._ip.symbolic:
                continue
            succ_addr = succ_state.addr

            # If we have set a bound for symbolic/concrete loops we want to handle it here
            if self.bound is not None:
                counts = succ_state.history.bbl_addrs.count(succ_addr)
                if counts > self.bound:
                    if self.bound_reached is not None:
                        # We want to pass self to modify the LocalLoopSeer state if needed
                        # Users can modify succ_state in the handler to implement their own logic
                        # or edit the state of LocalLoopSeer.
                        self.bound_reached(self, succ_state)
                    else:
                        # Remove the state from the successors object
                        # This state is going to be filtered by the self.filter function
                        self.cut_succs.append(succ_state)
        return succs



import angr
import claripy
import logging
from angr.errors import SimProcedureError

l = logging.getLogger(name=__name__)

# note: this does not handle skipping white space


class strtol(angr.SimProcedure):
    @staticmethod
    def strtol_inner(s, state, region, base, signed, read_length=None):
        """
        :param s: the string address/offset
        :param state: SimState
        :param region: memory, file, etc
        :param base: the base to use to interpret the number
        note: all numbers may start with +/- and base 16 may start with 0x
        :param signed: boolean, true means the result will be signed, otherwise unsigned
        :param read_length: int, the number of bytes parsed in strtol
        :return: expression, value, num_bytes
        the returned expression is a symbolic boolean indicating success, value will be set to 0 on failure
        value is the returned value (set to min/max on overflow)
        num_bytes is the number of bytes read in the string
        """

        # sanity check
        if base < 2 or base > 36:
            raise SimProcedureError("base should be in the range [2,36]")

        # order matters here since we will use an if then else tree, and -0x will have precedence over -
        prefixes = [b"-", b"+", b""]
        if base == 16:
            prefixes = [b"0x", b"-0x", b"+0x"] + prefixes

        cases = []
        conditions = []
        possible_num_bytes = []

        for prefix in prefixes:
            if read_length and read_length < len(prefix):
                continue
            condition, value, num_bytes = strtol._load_num_with_prefix(
                prefix, s, region, state, base, signed, read_length
            )
            conditions.append(condition)
            cases.append((condition, value))
            possible_num_bytes.append(num_bytes)

        # only one of the cases needed to match
        result = state.solver.ite_cases(cases[:-1], cases[-1][1])
        expression = state.solver.Or(*conditions)
        num_bytes = state.solver.ite_cases(zip(conditions, possible_num_bytes), 0)
        return expression, result, num_bytes

    @staticmethod
    def _load_num_with_prefix(prefix, addr, region, state, base, signed, read_length=None):
        """
        loads a number from addr, and returns a condition that addr must start with the prefix
        """
        length = len(prefix)
        read_length = (read_length - length) if read_length else None
        condition, value, num_bytes = strtol._string_to_int(addr + length, state, region, base, signed, read_length)

        # the prefix must match
        if len(prefix) > 0:
            loaded_prefix = region.load(addr, length)
            condition = state.solver.And(loaded_prefix == state.solver.BVV(prefix), condition)
        total_num_bytes = num_bytes + length

        # negatives
        if prefix.startswith(b"-"):
            value = state.solver.BVV(0, state.arch.bits) - value
        return condition, value, total_num_bytes

    @staticmethod
    def _string_to_int(s, state, region, base, signed, read_length=None):
        """
        reads values from s and generates the symbolic number that it would equal
        the first char is either a number in the given base, or the result is 0
        expression indicates whether or not it was successful
        """

        # if length wasn't provided, read the maximum bytes
        length = state.libc.max_strtol_len if read_length is None else read_length

        # expression whether or not it was valid at all
        expression, _ = strtol._char_to_val(region.load(s, 1), base)

        cases = []

        # to detect overflows we keep it in a larger bv and extract it at the end
        num_bits = min(state.arch.bits * 2, 128)
        current_val = state.solver.BVV(0, num_bits)
        num_bytes = state.solver.BVS("num_bytes", state.arch.bits)
        # constarints_num_bytes: a series of constraints of the form:
        # AND(<constraint on the string guaranteeing that the number of digits is n>, num_bytes == n)
        # these will be combined via OR and added to the state together num_bytes and the string
        constraints_num_bytes = []
        # conditions: one entry per byte loaded. contains the constraint that the byte is a parsable digit
        conditions = []

        # cutoff: whether the loop was broken with an unconvertable character
        cutoff = False
        # we need all the conditions to hold except the last one to have found a value
        for i in range(length):
            # begin reasoning about the currently indexed character
            char = region.load(s + i, 1)
            condition, value = strtol._char_to_val(char, base)

            # if it was the end we'll get the current val
            cases.append((num_bytes == i, current_val))

            # identify the constraints necessary to set num_bytes to the current value
            # the current char (i.e. the terminator if this is satisfied) should not be a char,
            # so `condition` should be false, plus all the previous conditions should be satisfied
            case_constraints = conditions + [state.solver.Not(condition), num_bytes == i]
            constraints_num_bytes.append(state.solver.And(*case_constraints))

            # break the loop early if no value past this is viable
            if condition.is_false():
                cutoff = True
                break

            # add the value and the condition
            current_val = current_val * base + value.zero_extend(num_bits - 8)
            conditions.append(condition)

        # if we ran out of bytes, we still need to add the case that every single byte was a digit
        if not cutoff:
            cases.append((num_bytes == length, current_val))
            case_constraints = conditions + [num_bytes == length]
            if read_length is None:
                # the loop broke because we hit angr's coded max. we need to assert that the following char is not
                # a digit in order for this case to generate correct models
                char = region.load(s + length, 1)
                condition, _ = strtol._char_to_val(char, base)
                case_constraints.append(state.solver.Not(condition))
            constraints_num_bytes.append(state.solver.And(*case_constraints))

        # only one of the constraints need to hold
        # since the constraints look like (num_bytes == 2 and the first 2 chars are valid, and the 3rd isn't)

        final_constraint = state.solver.Or(*constraints_num_bytes)
        if (
            final_constraint.op == "__eq__"
            and final_constraint.args[0] is num_bytes
            and not final_constraint.args[1].symbolic
        ):
            # CONCRETE CASE
            result = cases[state.solver.eval(final_constraint.args[1])][1]
            num_bytes = final_constraint.args[1]
        else:
            # symbolic case
            state.add_constraints(final_constraint)
            result = state.solver.ite_cases(cases, 0)

        # overflow check
        max_bits = state.arch.bits - 1 if signed else state.arch.bits
        max_val = 2**max_bits - 1
        result = state.solver.If(
            result < max_val,
            state.solver.Extract(state.arch.bits - 1, 0, result),
            state.solver.BVV(max_val, state.arch.bits),
        )

        return expression, result, num_bytes

    @staticmethod
    def _char_to_val(char, base):
        """
        converts a symbolic char to a number in the given base
        returns expression, result
        expression is a symbolic boolean indicating whether or not it was a valid number
        result is the value
        """
        cases = []
        # 0-9
        max_digit = claripy.BVV(b"9")
        min_digit = claripy.BVV(b"0")
        if base < 10:
            max_digit = claripy.BVV(ord("0") + base, 8)
        is_digit = claripy.And(char >= min_digit, char <= max_digit)
        # return early here so we don't add unnecessary statements
        if base <= 10:
            return is_digit, char - min_digit

        # handle alphabetic chars
        max_char_lower = claripy.BVV(ord("a") + base - 10 - 1, 8)
        max_char_upper = claripy.BVV(ord("A") + base - 10 - 1, 8)
        min_char_lower = claripy.BVV(ord("a"), 8)
        min_char_upper = claripy.BVV(ord("A"), 8)

        cases.append((is_digit, char - min_digit))
        is_alpha_lower = claripy.And(char >= min_char_lower, char <= max_char_lower)
        cases.append((is_alpha_lower, char - min_char_lower + 10))
        is_alpha_upper = claripy.And(char >= min_char_upper, char <= max_char_upper)
        cases.append((is_alpha_upper, char - min_char_upper + 10))

        expression = claripy.Or(is_digit, is_alpha_lower, is_alpha_upper)
        # use the last case as the default, the expression will encode whether or not it's satisfiable
        result = claripy.ite_cases(cases[:-1], cases[-1][1])

        return expression, result

    def run(self, nptr, endptr, base):  # pylint: disable=arguments-differ
        if self.state.solver.symbolic(base):
            l.warning("Concretizing symbolic base in strtol")
            base_concrete = self.state.solver.eval(base)
            self.state.add_constraints(base == base_concrete)

        base = self.state.solver.eval(base)

        if base == 0:
            # in this case the base is 16 if it starts with 0x, 8 if it starts with 0, 10 otherwise
            # here's the possibilities
            base_16_pred = self.state.solver.Or(
                self.state.memory.load(nptr, 2) == self.state.solver.BVV(b"0x"),
                self.state.memory.load(nptr, 3) == self.state.solver.BVV(b"+0x"),
                self.state.memory.load(nptr, 3) == self.state.solver.BVV(b"-0x"),
            )
            base_8_pred = self.state.solver.And(
                self.state.solver.Or(
                    self.state.memory.load(nptr, 1) == self.state.solver.BVV(b"0"),
                    self.state.memory.load(nptr, 2) == self.state.solver.BVV(b"+0"),
                    self.state.memory.load(nptr, 2) == self.state.solver.BVV(b"-0"),
                ),
                self.state.solver.Not(base_16_pred),
            )
            base_10_pred = self.state.solver.And(
                self.state.solver.Not(base_16_pred), self.state.solver.Not(base_8_pred)
            )
            expressions = []
            values = []
            num_bytes_arr = []

            # read a string to long for each possibility
            pred_base = zip([base_16_pred, base_10_pred, base_8_pred], [16, 10, 8])
            for pred, sub_base in pred_base:
                expression, value, num_bytes = self.strtol_inner(nptr, self.state, self.state.memory, sub_base, True)
                expressions.append(self.state.solver.And(expression, pred))
                values.append(value)
                num_bytes_arr.append(num_bytes)

            # we would return the Or(expressions) as the indicator whether or not it succeeded, but it's not needed
            # for strtol
            # expression = self.state.solver.Or(expressions)
            value = self.state.solver.ite_cases(zip(expressions, values), 0)
            num_bytes = self.state.solver.ite_cases(zip(expressions, num_bytes_arr), 0)

            self.state.memory.store(
                endptr, nptr + num_bytes, condition=(endptr != 0), endness=self.state.arch.memory_endness
            )

            return value

        else:
            expression, value, num_bytes = self.strtol_inner(nptr, self.state, self.state.memory, base, True)
            self.state.memory.store(
                endptr, nptr + num_bytes, condition=(endptr != 0), endness=self.state.arch.memory_endness
            )
            return self.state.solver.If(expression, value, 0)



#!/usr/bin/env python3
from unittest import main, TestCase

import claripy

from angr.knowledge_plugins.key_definitions.environment import Environment
from angr.knowledge_plugins.key_definitions.undefined import UNDEFINED


class TestEnvironment(TestCase):
    def setUp(self):
        self.pointer_length = 8
        self.pointer = 0x7FFFFFFFF00 - 0xFF

    def test_get_returns_the_pointer_associated_with_an_existing_environment_variable(self):
        name = "variable_name"
        data = {claripy.BVV(self.pointer, self.pointer_length)}
        environment = Environment(environment={name: data})

        variable_names = {name}

        result_values, _ = environment.get(variable_names)
        self.assertEqual(result_values, data)

    def test_get_returns_a_dataset_with_UNDEFINED_when_the_environment_variable_is_not_set(self):
        environment = Environment(environment={})

        name = "variable_name"
        variable_names = {name}

        result_values, _ = environment.get(variable_names)
        self.assertEqual(result_values, {UNDEFINED})

    def test_get_returns_values_associated_with_all_possible_given_names(self):
        environment_data = {
            "variable_name": {claripy.BVV(self.pointer, self.pointer_length)},
            UNDEFINED: {UNDEFINED},
        }
        environment = Environment(environment=environment_data)

        variable_names = {"variable_name", UNDEFINED}
        expected_result = {claripy.BVV(self.pointer, self.pointer_length), UNDEFINED}

        result_values, _ = environment.get(variable_names)
        self.assertEqual(result_values, expected_result)

    def test_get_fails_when_called_with_wrong_type_of_name(self):
        environment = Environment(environment={})
        variable_names = {claripy.BVV(0x42, 8 * 8)}

        self.assertRaises(TypeError, environment.get, variable_names)

    def test_get_result_tells_if_environment_knows_about_all_the_variables(self):
        name = "variable_name"
        data = {claripy.BVV(self.pointer, self.pointer_length)}
        environment = Environment(environment={name: data})

        name = "variable_name"
        variable_names = {name}

        _, has_unknown = environment.get(variable_names)
        self.assertFalse(has_unknown)

    def test_get_result_tells_if_environment_does_not_know_about_one_of_the_variable(self):
        name = "variable_name"
        data = {claripy.BVV(b"value\x00", 6 * 8)}
        environment = Environment(environment={name: data})

        name = "variable_name"
        variable_names = {name, "unknown_name"}

        _, has_unknown = environment.get(variable_names)
        self.assertTrue(has_unknown)

    def test_set_sets_associates_data_to_a_given_variable_name(self):
        environment = Environment(environment={})

        data = {claripy.BVV(b"value\x00", 6 * 8)}
        environment.set("variable_name", data)

        # Testing internal attribute which is not ideal, but I'am trying to not rely on any other behavior
        # (e.g. without using `get`).
        self.assertDictEqual(environment._environment, {"variable_name": data})

    def test_set_sets_can_associate_data_to_UNDEFINED(self):
        environment = Environment(environment={})

        data = {claripy.BVV(b"value\x00", 6 * 8)}
        environment.set(UNDEFINED, data)

        # Testing internal attribute which is not ideal, but I'am trying to not rely on any other behavior
        # (e.g. without using `get`).
        self.assertDictEqual(environment._environment, {UNDEFINED: data})

    def test_set_fails_when_wvariable_name_is_of_wrong_type(self):
        environment = Environment(environment={})

        data = {claripy.BVV(b"value\x00", 6 * 8)}

        self.assertRaises(TypeError, environment.set, 0x42, data)

    def test_merge_fails_when_merging_with_a_non_Environment_instance(self):
        environment = Environment(environment={"variable_name": {claripy.BVV(b"value1\x00", 7 * 8)}})
        other_environment = 0x42

        self.assertRaises(TypeError, environment.merge, other_environment)

    def test_merge_two_environments_merge_data_associated_with_each_variable(self):
        first = "variable_name"
        second = "other_variable_name"
        environment = Environment(environment={first: {claripy.BVV(b"value1\x00", 7 * 8)}})
        other_environment = Environment(environment={first: {claripy.BVV(b"value2\x00", 7 * 8)}, second: {UNDEFINED}})

        expected_environment = Environment(
            environment={
                first: {claripy.BVV(b"value1\x00", 7 * 8), claripy.BVV(b"value2\x00", 7 * 8)},
                second: {UNDEFINED},
            }
        )

        env, _ = environment.merge(other_environment)
        self.assertEqual(env, expected_environment)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,no-self-use,line-too-long

from collections import OrderedDict
import ctypes
import unittest

from archinfo import Endness
import claripy

import angr
from angr import SimState
from angr.sim_type import register_types, parse_types, SimStruct, SimTypeNumOffset


class TestMemView(unittest.TestCase):
    def test_simple_concrete(self):
        s = SimState(arch="AMD64")
        addr = 0xBA5E0

        def check_read(val):
            assert s.solver.eval(s.memory.load(addr, 8, endness=Endness.LE), cast_to=int) == val

            assert s.mem[addr].char.concrete == chr(val & 0xFF).encode()
            assert s.mem[addr].byte.concrete == val & 0xFF

            assert s.mem[addr].int16_t.concrete == ctypes.c_int16(val & 0xFFFF).value
            assert s.mem[addr].uint16_t.concrete == val & 0xFFFF

            assert s.mem[addr].qword.concrete == val

        s.memory.store(addr, claripy.BVV(0x11223344AABBCC7D, 64), endness=Endness.LE)
        check_read(0x11223344AABBCC7D)

        # test storing
        s.mem[addr].uint16_t = 0xEF6D
        check_read(0x11223344AABBEF6D)

    def test_string_concrete(self):
        s = SimState(arch="AMD64")
        addr = 0xBA5E0

        def check_read(val):
            assert s.solver.eval(s.memory.load(addr, len(val)), cast_to=bytes) == val
            assert s.solver.eval(s.memory.load(addr + len(val), 1), cast_to=int) == 0

            assert s.mem[addr].string.concrete == val

        s.memory.store(addr, b"a string!\0")
        check_read(b"a string!")

        # not supported yet
        # s.mem[addr].string = "shorter"
        # check_read(b"shorter")

        # s.mem[addr].string = "a longer string"
        # check_read(b"a longer string")

    def test_array_concrete(self):
        s = SimState(arch="AMD64")
        addr = 0xBA5E0

        s.memory.store(addr, claripy.BVV(0x1, 32), endness=Endness.LE)
        s.memory.store(addr + 4, claripy.BVV(0x2, 32), endness=Endness.LE)
        s.memory.store(addr + 8, claripy.BVV(0x3, 32), endness=Endness.LE)
        s.memory.store(addr + 12, claripy.BVV(0x4, 32), endness=Endness.LE)
        s.memory.store(addr + 16, claripy.BVV(0x5, 32), endness=Endness.LE)

        assert s.mem[addr].dword.array(5).concrete == [0x1, 0x2, 0x3, 0x4, 0x5]
        assert s.mem[addr].dword.array(5)[2].concrete == 0x3
        assert s.mem[addr].qword.array(2).concrete == [
            0x0000000200000001,
            0x0000000400000003,
        ]
        assert s.mem[addr].dword.array(2).array(2).concrete == [[0x1, 0x2], [0x3, 0x4]]

        s.mem[addr].dword.array(5)[3] = 10
        assert s.solver.eval(s.memory.load(addr + 12, 4, endness=Endness.LE), cast_to=int) == 10

        s.mem[addr].dword.array(5).store([20, 2, 3, 4, 5])
        assert s.mem[addr].dword.array(4).concrete == [20, 2, 3, 4]

        s.mem[addr].dword.array(2).array(2).store([[1, 2], [4, 3]])
        assert s.mem[addr].dword.array(4).concrete == [1, 2, 4, 3]

    def test_pointer_concrete(self):
        s = SimState(arch="AMD64")
        addr = 0xBA5E0
        ptraddr = 0xCD0

        s.memory.store(ptraddr, claripy.BVV(addr, 64), endness=Endness.LE)
        s.memory.store(addr, b"abcdef\0")

        assert s.mem[ptraddr].deref.string.concrete == b"abcdef"
        s.mem[ptraddr].deref.dword = 123954
        assert s.solver.eval(s.memory.load(addr, 4, endness=Endness.LE), cast_to=int) == 123954
        assert s.mem[ptraddr].deref.dword.concrete == 123954

    def test_structs(self):
        s = SimState(arch="AMD64")

        register_types(
            parse_types(
                """
    struct test_structs {
      int a;
      long b;
    };
    """
            )
        )

        s.memory.store(0x8000, bytes(16))
        s.mem[0x8000].struct.test_structs = {"a": 10, "b": 20}
        assert s.mem[0x8000].struct.test_structs.a.concrete == 10
        assert s.solver.eval(s.memory.load(0x8000, 16), cast_to=bytes) == bytes.fromhex(
            "0a000000000000001400000000000000"
        )

    def test_struct_bitfield_simple(self):
        """
        Tests if a struct with bitfields like
        struct {
            uint32_t a:8, b:1, c:23;
        }
        can be used with a memview
        :return:
        """
        state = SimState(arch="AMD64")
        register_types(
            SimStruct(
                name="bitfield_struct",
                pack=True,
                fields=OrderedDict(
                    [
                        ("a", SimTypeNumOffset(8, signed=False)),
                        ("b", SimTypeNumOffset(1, signed=False)),
                        ("c", SimTypeNumOffset(23, signed=False)),
                    ]
                ),
            )
        )

        data = [
            (b"\x0e\x02\x00\x00", (14, 0, 1)),
            (b"\x14T\x00\x00", (20, 0, 42)),
            (b"\x04\n\x01\x00", (4, 0, 133)),
            (b"\x04j\x01\x00", (4, 0, 181)),
            (b"\x04\xa2\x01\x00", (4, 0, 209)),
            (b"\x04\xf4\x01\x00", (4, 0, 250)),
            (b"\x04\\\x02\x00", (4, 0, 302)),
            (b"\x04\x98\x02\x00", (4, 0, 332)),
            (b"\x04\xe0\x02\x00", (4, 0, 368)),
            (b"\x04\x1e\x03\x00", (4, 0, 399)),
        ]
        state.memory.store(
            0x8000,
            b"\x0e\x02\x00\x00"
            b"\x14T\x00\x00"
            b"\x04\n\x01\x00"
            b"\x04j\x01\x00"
            b"\x04\xa2\x01\x00"
            b"\x04\xf4\x01\x00"
            b"\x04\\\x02\x00"
            b"\x04\x98\x02\x00"
            b"\x04\xe0\x02\x00"
            b"\x04\x1e\x03\x00",
        )
        view = state.mem[0x8000].struct.bitfield_struct.array(5)
        for idx, (b, result) in enumerate(data):
            v = view[idx]
            s = v.concrete
            assert s.a == result[0], f"Field a was {s.a}, expected {result[0]}, from bytes {b}"
            assert v.a.concrete == result[0], f"Field a was {v.a.concrete}, expected {result[0]}, from bytes {b}"

            assert s.b == result[1], f"Field b was {s.b}, expected {result[1]}, from bytes {b}"
            assert v.b.concrete == result[1], f"Field b was {s.b}, expected {result[1]}, from bytes {b}"

            assert s.c == result[2], f"Field c was {s.c}, expected {result[2]}, from bytes {b}"
            assert v.c.concrete == result[2], f"Field c was {v.c.concrete}, expected {result[2]}, from bytes {b}"

    def test_struct_bitfield_complex(self):
        bitfield_struct2 = angr.types.parse_type(
            """struct bitfield_struct2
        {
            uint64_t    target    : 36,
                        high8     :  8,
                        reserved  :  7,
                        next      : 12,
                        bind      :  1;
        }"""
        )

        angr.types.register_types(bitfield_struct2)
        state = SimState(arch="AMD64")
        state.memory.store(0x1000, b"\xb3\xc7\xe9|\xad\xd7\xee$")  # store some random data
        struct = state.mem[0x1000].struct.bitfield_struct2.concrete
        assert struct.target == 0xD7CE9C7B3
        assert struct.high8 == 0x7A
        assert struct.next == 0x49D
        assert struct.bind == 0


if __name__ == "__main__":
    unittest.main()



from .base import (
    BaseStructuredCodeGenerator,
    InstructionMapping,
    InstructionMappingElement,
    PositionMappingElement,
    PositionMapping,
)
from .c import CStructuredCodeGenerator
from .dwarf_import import ImportSourceCode
from .dummy import DummyStructuredCodeGenerator



# pylint:disable=abstract-method,ungrouped-imports

from typing import Set, List, Optional, TYPE_CHECKING
import logging

import pyvex

from angr.analyses import ForwardAnalysis, visitors
from ..utils.constants import is_alignment_mask
from ..analyses import AnalysesHub
from ..knowledge_plugins import Function
from ..block import BlockNode
from ..errors import SimTranslationError
from .analysis import Analysis

try:
    import pypcode
    from angr.engines import pcode
except ImportError:
    pypcode = None

if TYPE_CHECKING:
    from angr.block import Block

_l = logging.getLogger(name=__name__)


class BottomType:
    """
    The bottom value for register values.
    """

    def __repr__(self):
        return "<Bottom>"


TOP = None
BOTTOM = BottomType()


class Constant:
    """
    Represents a constant value.
    """

    __slots__ = ("val",)

    def __init__(self, val):
        self.val = val

    def __eq__(self, other):
        if type(other) is Constant or isinstance(other, Constant):
            return self.val == other.val
        return False

    def __hash__(self):
        return hash((Constant, self.val))

    def __repr__(self):
        return repr(self.val)

    def __add__(self, other):
        if type(self) is type(other):
            return Constant(self.val + other.val)
        else:
            return other + self

    def __sub__(self, other):
        if type(self) is type(other):
            return Constant(self.val + other.val)
        else:
            raise CouldNotResolveException


class Register:
    """
    Represent a register.
    """

    __slots__ = ("offset", "bitlen")

    def __init__(self, offset, bitlen):
        self.offset = offset
        self.bitlen = bitlen

    def __hash__(self):
        return hash((Register, self.offset))

    def __eq__(self, other):
        if type(other) is Register or isinstance(other, Register):
            return self.offset == other.offset
        return False

    def __repr__(self):
        return str(self.offset)


class OffsetVal:
    """
    Represent a value with an offset added.
    """

    __slots__ = (
        "_reg",
        "_offset",
    )

    def __init__(self, reg, offset):
        self._reg = reg
        self._offset = offset

    @property
    def reg(self):
        return self._reg

    @property
    def offset(self):
        return self._offset

    def __add__(self, other):
        if type(other) is Constant:
            return OffsetVal(self._reg, (self._offset + other.val) & (2**self.reg.bitlen - 1))
        else:
            raise CouldNotResolveException

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if type(other) is Constant:
            return OffsetVal(self._reg, self._offset - other.val & (2**self.reg.bitlen - 1))
        else:
            raise CouldNotResolveException

    def __rsub__(self, other):
        raise CouldNotResolveException

    def __eq__(self, other):
        if type(other) is OffsetVal or isinstance(other, OffsetVal):
            return self.reg == other.reg and self.offset == other.offset
        return False

    def __hash__(self):
        return hash((type(self), self._reg, self._offset))

    def __repr__(self):
        return f"reg({self.reg}){(self.offset - 2**self.reg.bitlen) if self.offset != 0 else 0:+}"


class FrozenStackPointerTrackerState:
    """
    Abstract state for StackPointerTracker analysis with registers and memory values being in frozensets.
    """

    __slots__ = "regs", "memory", "is_tracking_memory"

    def __init__(self, regs, memory, is_tracking_memory):
        self.regs = regs
        self.memory = memory
        self.is_tracking_memory = is_tracking_memory

    def unfreeze(self):
        return StackPointerTrackerState(dict(self.regs), dict(self.memory), self.is_tracking_memory)

    def __hash__(self):
        if self.is_tracking_memory:
            return hash((FrozenStackPointerTrackerState, self.regs, self.memory, self.is_tracking_memory))
        else:
            return hash((FrozenStackPointerTrackerState, self.regs, self.is_tracking_memory))

    def merge(self, other):
        return self.unfreeze().merge(other.unfreeze()).freeze()

    def __eq__(self, other):
        if type(other) is FrozenStackPointerTrackerState or isinstance(other, FrozenStackPointerTrackerState):
            cond1 = self.regs == other.regs and self.is_tracking_memory == other.is_tracking_memory
            if self.is_tracking_memory:
                cond1 &= self.memory == other.memory
            return cond1
        return False


class StackPointerTrackerState:
    """
    Abstract state for StackPointerTracker analysis.
    """

    __slots__ = "regs", "memory", "is_tracking_memory"

    def __init__(self, regs, memory, is_tracking_memory):
        self.regs = regs
        if is_tracking_memory:
            self.memory = memory
        else:
            self.memory = {}
        self.is_tracking_memory = is_tracking_memory

    def give_up_on_memory_tracking(self):
        self.memory = {}
        self.is_tracking_memory = False

    def store(self, addr, val):
        # strong update
        if self.is_tracking_memory and val is not None and addr is not None:
            self.memory[addr] = val

    def load(self, addr):
        if not self.is_tracking_memory:
            return TOP
        try:
            val = self.memory[addr]
            if val is not TOP:
                return val
        except KeyError:
            pass
        raise CouldNotResolveException

    def get(self, reg):
        try:
            val = self.regs[reg]
            if val is not TOP:
                return val
        except KeyError:
            pass
        raise CouldNotResolveException

    def put(self, reg, val):
        # strong update, but we only update values for registers that are already in self.regs and ignore all other
        # registers. obviously, self.regs should be initialized with registers that should be considered during
        # tracking,
        if reg in self.regs:
            self.regs[reg] = val

    def copy(self):
        return StackPointerTrackerState(self.regs.copy(), self.memory.copy(), self.is_tracking_memory)

    def freeze(self):
        return FrozenStackPointerTrackerState(
            frozenset(self.regs.items()), frozenset(self.memory.items()), self.is_tracking_memory
        )

    def __eq__(self, other):
        if type(other) is StackPointerTrackerState or isinstance(other, StackPointerTrackerState):
            cond1 = self.regs == other.regs and self.is_tracking_memory == other.is_tracking_memory
            if self.is_tracking_memory:
                cond1 &= self.memory == other.memory
            return cond1
        return False

    def __hash__(self):
        if self.is_tracking_memory:
            return hash((StackPointerTrackerState, self.regs, self.memory, self.is_tracking_memory))
        else:
            return hash((StackPointerTrackerState, self.regs, self.is_tracking_memory))

    def merge(self, other):
        return StackPointerTrackerState(
            regs=_dict_merge(self.regs, other.regs),
            memory=_dict_merge(self.memory, other.memory),
            is_tracking_memory=self.is_tracking_memory and other.is_tracking_memory,
        )


def _dict_merge(d1, d2):
    all_keys = set(d1.keys()) | set(d2.keys())
    merged = {}
    for k in all_keys:
        if k not in d1 or d1[k] is TOP:
            merged[k] = TOP
        elif k not in d2 or d2[k] is TOP:
            merged[k] = TOP
        elif d1[k] is BOTTOM:
            merged[k] = d2[k]
        elif d2[k] is BOTTOM:
            merged[k] = d1[k]
        elif d1[k] == d2[k]:
            merged[k] = d1[k]
        else:  # d1[k] != d2[k]
            merged[k] = TOP
    return merged


class CouldNotResolveException(Exception):
    """
    An exception used in StackPointerTracker analysis to represent internal resolving failures.
    """


class StackPointerTracker(Analysis, ForwardAnalysis):
    """
    Track the offset of stack pointer at the end of each basic block of a function.
    """

    def __init__(
        self, func: Optional[Function], reg_offsets: Set[int], block: Optional["Block"] = None, track_memory=True
    ):
        if func is not None:
            if not func.normalized:
                # Make a copy before normalizing the function
                func = func.copy()
                func.normalize()
            graph_visitor = visitors.FunctionGraphVisitor(func)
        elif block is not None:
            graph_visitor = visitors.SingleNodeGraphVisitor(block)
        else:
            raise ValueError("StackPointerTracker must work on either a function or a single block.")

        super().__init__(order_jobs=False, allow_merging=True, allow_widening=track_memory, graph_visitor=graph_visitor)

        self.track_mem = track_memory
        self._func = func
        self.reg_offsets = reg_offsets
        self.states = {}
        self._blocks = {}

        _l.debug("Running on function %r", self._func)
        self._analyze()

    def _state_for(self, addr, pre_or_post):
        if addr not in self.states:
            return None

        addr_map = self.states[addr]
        if pre_or_post not in addr_map:
            return None

        return addr_map[pre_or_post]

    def _offset_for(self, addr, pre_or_post, reg):
        try:
            s = self._state_for(addr, pre_or_post)
            if s is None:
                return TOP
            regval = dict(s.regs)[reg]
        except KeyError:
            return TOP
        if regval is TOP or type(regval) is Constant:
            return TOP
        elif regval is BOTTOM:
            # we don't really know what it should be. return TOP instead.
            return TOP
        else:
            return regval.offset

    def offset_after(self, addr, reg):
        return self._offset_for(addr, "post", reg)

    def offset_before(self, addr, reg):
        return self._offset_for(addr, "pre", reg)

    def offset_after_block(self, block_addr, reg):
        if block_addr not in self._blocks:
            return TOP
        instr_addrs = self._blocks[block_addr].instruction_addrs
        if len(instr_addrs) == 0:
            return TOP
        else:
            return self.offset_after(instr_addrs[-1], reg)

    def offset_before_block(self, block_addr, reg):
        if block_addr not in self._blocks:
            return TOP
        instr_addrs = self._blocks[block_addr].instruction_addrs
        if len(instr_addrs) == 0:
            return TOP
        else:
            return self.offset_before(instr_addrs[0], reg)

    @property
    def inconsistent(self):
        return any(self.inconsistent_for(r) for r in self.reg_offsets)

    def inconsistent_for(self, reg):
        for endpoint in self._func.endpoints:
            if self.offset_after_block(endpoint.addr, reg) is TOP:
                return True
        return False

    #
    # Overridable methods
    #

    def _pre_analysis(self):
        pass

    def _intra_analysis(self):
        pass

    def _post_analysis(self):
        pass

    def _get_register(self, offset):
        name = self.project.arch.register_names[offset]
        size = self.project.arch.registers[name][1]
        return Register(offset, size * self.project.arch.byte_width)

    def _initial_abstract_state(self, node: BlockNode):
        if self._func is None:
            # in single-block mode, at the beginning of the block, we set each tracking register to their initial values
            initial_regs = {r: OffsetVal(self._get_register(r), 0) for r in self.reg_offsets}
        else:
            # function mode
            if node.addr == self._func.addr:
                # at the beginning of the function, we set each tracking register to their "initial values"
                initial_regs = {r: OffsetVal(self._get_register(r), 0) for r in self.reg_offsets}
            else:
                # if we are requesting initial states for blocks that are not the starting point of this function, we
                # are probably dealing with dangling blocks (those without a predecessor due to CFG recovery failures).
                # Setting register values to fresh ones will cause problems down the line when merging with normal
                # register values happen. therefore, we set their values to BOTTOM. these BOTTOMs will be replaced once
                # a merge with normal blocks happen.
                initial_regs = {r: BOTTOM for r in self.reg_offsets}

        return StackPointerTrackerState(regs=initial_regs, memory={}, is_tracking_memory=self.track_mem).freeze()

    def _set_state(self, addr, new_val, pre_or_post):
        previous_val = self._state_for(addr, pre_or_post)
        if previous_val is not None:
            new_val = previous_val.merge(new_val)
        if addr not in self.states:
            self.states[addr] = {}
        self.states[addr][pre_or_post] = new_val

    def _set_post_state(self, addr, new_val):
        self._set_state(addr, new_val, "post")

    def _set_pre_state(self, addr, new_val):
        self._set_state(addr, new_val, "pre")

    def _run_on_node(self, node: BlockNode, state):
        block = self.project.factory.block(node.addr, size=node.size)
        self._blocks[node.addr] = block

        state = state.unfreeze()
        _l.debug("START:       Running on block at %x", node.addr)
        _l.debug("Regs: %s", state.regs)
        _l.debug("Mem: %s", state.memory)
        curr_stmt_start_addr = None

        vex_block = None
        try:
            vex_block = block.vex
        except SimTranslationError:
            pass

        if vex_block is not None:
            if isinstance(vex_block, pyvex.IRSB):
                curr_stmt_start_addr = self._process_vex_irsb(node, vex_block, state)
            elif pypcode is not None and isinstance(vex_block, pcode.lifter.IRSB):
                curr_stmt_start_addr = self._process_pcode_irsb(node, vex_block, state)
            else:
                raise NotImplementedError(f"Unsupported block type {type(vex_block)}")

        if curr_stmt_start_addr is not None:
            self._set_post_state(curr_stmt_start_addr, state.freeze())

        _l.debug("FINISH:      After running on block at %x", node.addr)
        _l.debug("Regs: %s", state.regs)
        _l.debug("Mem: %s", state.memory)

        output_state = state.freeze()
        return None, output_state

    def _process_vex_irsb(self, node, vex_block: pyvex.IRSB, state: StackPointerTrackerState) -> int:
        tmps = {}
        curr_stmt_start_addr = None

        def _resolve_expr(expr):
            if type(expr) is pyvex.IRExpr.Binop:
                arg0, arg1 = expr.args
                if expr.op.startswith("Iop_Add"):
                    arg0_expr = _resolve_expr(arg0)
                    if arg0_expr is None:
                        raise CouldNotResolveException()
                    if arg0_expr is BOTTOM:
                        return BOTTOM
                    arg1_expr = _resolve_expr(arg1)
                    if arg1_expr is None:
                        raise CouldNotResolveException()
                    if arg1_expr is BOTTOM:
                        return BOTTOM
                    return arg0_expr + arg1_expr
                elif expr.op.startswith("Iop_Sub"):
                    arg0_expr = _resolve_expr(arg0)
                    if arg0_expr is None:
                        raise CouldNotResolveException()
                    if arg0_expr is BOTTOM:
                        return BOTTOM
                    arg1_expr = _resolve_expr(arg1)
                    if arg1_expr is None:
                        raise CouldNotResolveException()
                    if arg1_expr is BOTTOM:
                        return BOTTOM
                    return arg0_expr - arg1_expr
                elif expr.op.startswith("Iop_And"):
                    # handle stack pointer alignments
                    arg0_expr = _resolve_expr(arg0)
                    arg1_expr = _resolve_expr(arg1)
                    if (
                        isinstance(arg1_expr, (Register, OffsetVal))
                        and isinstance(arg0_expr, Constant)
                        and is_alignment_mask(arg0_expr.val)
                    ):
                        return arg1_expr
                    if (
                        isinstance(arg0_expr, (Register, OffsetVal))
                        and isinstance(arg1_expr, Constant)
                        and is_alignment_mask(arg1_expr.val)
                    ):
                        return arg0_expr
                    raise CouldNotResolveException()
            elif type(expr) is pyvex.IRExpr.RdTmp and expr.tmp in tmps and tmps[expr.tmp] is not None:
                return tmps[expr.tmp]
            elif type(expr) is pyvex.IRExpr.Const:
                return Constant(expr.con.value)
            elif type(expr) is pyvex.IRExpr.Get:
                return state.get(expr.offset)
            elif self.track_mem and type(expr) is pyvex.IRExpr.Load:
                return state.load(_resolve_expr(expr.addr))
            raise CouldNotResolveException()

        def resolve_expr(expr):
            try:
                return _resolve_expr(expr)
            except CouldNotResolveException:
                return TOP

        def resolve_stmt(stmt):
            if type(stmt) is pyvex.IRStmt.WrTmp:
                tmps[stmt.tmp] = resolve_expr(stmt.data)
            elif self.track_mem and type(stmt) is pyvex.IRStmt.Store:
                state.store(resolve_expr(stmt.addr), resolve_expr(stmt.data))
            elif type(stmt) is pyvex.IRStmt.Put:
                if exit_observed and stmt.offset == self.project.arch.sp_offset:
                    return
                state.put(stmt.offset, resolve_expr(stmt.data))
            else:
                raise CouldNotResolveException()

        exit_observed = False
        for stmt in vex_block.statements:
            if type(stmt) is pyvex.IRStmt.IMark:
                if curr_stmt_start_addr is not None:
                    # we've reached a new instruction. Time to store the post state
                    self._set_post_state(curr_stmt_start_addr, state.freeze())
                curr_stmt_start_addr = stmt.addr + stmt.delta
                self._set_pre_state(curr_stmt_start_addr, state.freeze())
            elif (
                type(stmt) is pyvex.IRStmt.Exit
                and curr_stmt_start_addr in vex_block.instruction_addresses
                and vex_block.instruction_addresses.index(curr_stmt_start_addr) == vex_block.instructions - 1
            ):
                exit_observed = True
            else:
                try:
                    resolve_stmt(stmt)
                except CouldNotResolveException:
                    pass

        # stack pointer adjustment
        if self.project.arch.sp_offset in self.reg_offsets and vex_block.jumpkind == "Ijk_Call":
            if self.project.arch.call_pushes_ret:
                # pop the return address on the stack
                try:
                    v = state.get(self.project.arch.sp_offset)
                    if v is BOTTOM:
                        incremented = BOTTOM
                    else:
                        incremented = v + Constant(self.project.arch.bytes)
                    state.put(self.project.arch.sp_offset, incremented)
                except CouldNotResolveException:
                    pass
            # who are we calling?
            callees = [] if self._func is None else self._find_callees(node)
            if callees:
                callee_cleanups = [
                    callee
                    for callee in callees
                    if callee.calling_convention is not None and callee.calling_convention.CALLEE_CLEANUP
                ]
                if callee_cleanups:
                    # found callee clean-up cases...
                    try:
                        v = state.get(self.project.arch.sp_offset)
                        if v is BOTTOM:
                            incremented = BOTTOM
                        elif callee_cleanups[0].prototype is not None:
                            num_args = len(callee_cleanups[0].prototype.args)
                            incremented = v + Constant(self.project.arch.bytes * num_args)
                        state.put(self.project.arch.sp_offset, incremented)
                    except CouldNotResolveException:
                        pass

        return curr_stmt_start_addr

    def _process_pcode_irsb(self, node, pcode_irsb: "pcode.lifter.IRSB", state: StackPointerTrackerState) -> int:
        unique = {}
        curr_stmt_start_addr = None

        def _resolve_expr(varnode: "pypcode.Varnode"):
            if varnode.space.name == "register":
                return state.get(varnode.offset)
            elif varnode.space.name == "unique":
                key = (varnode.offset, varnode.size)
                if key not in unique:
                    raise CouldNotResolveException()
                return unique[key]
            elif varnode.space.name == "const":
                return Constant(varnode.offset)
            else:
                raise CouldNotResolveException()

        def resolve_expr(varnode: "pypcode.Varnode"):
            try:
                return _resolve_expr(varnode)
            except CouldNotResolveException:
                return TOP

        def resolve_op(op: "pypcode.PcodeOp"):
            if op.opcode == pypcode.OpCode.INT_ADD and len(op.inputs) == 2:
                input0, input1 = op.inputs
                input0_v = resolve_expr(input0)
                input1_v = resolve_expr(input1)
                if isinstance(input0_v, (Register, OffsetVal)) and isinstance(input1_v, Constant):
                    v = input0_v + input1_v
                else:
                    raise CouldNotResolveException()
            elif op.opcode == pypcode.OpCode.COPY:
                v = resolve_expr(op.inputs[0])
            else:
                # unsupported opcode
                raise CouldNotResolveException()

            # write the output
            if op.output.space.name == "unique":
                offset, size = op.output.offset, op.output.size
                unique[(offset, size)] = v
            elif op.output.space.name == "register":
                state.put(op.output.offset, v)
            else:
                raise CouldNotResolveException()

        for instr in pcode_irsb._instructions:
            if curr_stmt_start_addr is not None:
                # we've reached a new instruction. Time to store the post state
                self._set_post_state(curr_stmt_start_addr, state.freeze())
            curr_stmt_start_addr = instr.address.offset
            self._set_pre_state(curr_stmt_start_addr, state.freeze())

            for op in instr.ops:
                op: "pypcode.PcodeOp"
                try:
                    resolve_op(op)
                except CouldNotResolveException:
                    pass

        # examine the last instruction and determine if this is a call instruction
        is_call = False
        if pcode_irsb._instructions:
            last_instr = pcode_irsb._instructions[-1]
            if last_instr.ops:
                last_op = last_instr.ops[-1]
                if last_op.opcode == pypcode.OpCode.CALL:
                    is_call = True

        # stack pointer adjustment
        if self.project.arch.sp_offset in self.reg_offsets and is_call:
            if self.project.arch.call_pushes_ret:
                # pop the return address on the stack
                try:
                    v = state.get(self.project.arch.sp_offset)
                    if v is BOTTOM:
                        incremented = BOTTOM
                    else:
                        incremented = v + Constant(self.project.arch.bytes)
                    state.put(self.project.arch.sp_offset, incremented)
                except CouldNotResolveException:
                    pass
            # who are we calling?
            callees = self._find_callees(node)
            if callees:
                callee_cleanups = [
                    callee
                    for callee in callees
                    if callee.calling_convention is not None and callee.calling_convention.CALLEE_CLEANUP
                ]
                if callee_cleanups:
                    # found callee clean-up cases...
                    try:
                        v = state.get(self.project.arch.sp_offset)
                        if v is BOTTOM:
                            incremented = BOTTOM
                        elif callee_cleanups[0].prototype is not None:
                            num_args = len(callee_cleanups[0].prototype.args)
                            incremented = v + Constant(self.project.arch.bytes * num_args)
                        state.put(self.project.arch.sp_offset, incremented)
                    except CouldNotResolveException:
                        pass

        return curr_stmt_start_addr

    def _widen_states(self, *states):
        assert len(states) == 2
        merged, _ = self._merge_states(None, *states)
        if len(merged.memory) > 5:
            _l.info("Encountered too many memory writes in stack pointer tracking. Abandoning memory tracking.")
            merged = merged.unfreeze().give_up_on_memory_tracking().freeze()
        return merged

    def _merge_states(self, node, *states: StackPointerTrackerState):
        merged_state = states[0]
        for other in states[1:]:
            merged_state = merged_state.merge(other)
        return merged_state, merged_state == states[0]

    def _find_callees(self, node) -> List[Function]:
        callees: List[Function] = []
        for _, dst, data in self._func.transition_graph.out_edges(node, data=True):
            if data.get("type") == "call":
                if isinstance(dst, Function):
                    callees.append(dst)
        return callees


AnalysesHub.register_default("StackPointerTracker", StackPointerTracker)



#!/usr/bin/env python3
# Disable some pylint warnings: no-self-use, missing-docstring
# pylint: disable=R0201,C0111,bad-builtin

import os
import pickle
from unittest import TestCase, main

import ailment
import claripy

import angr
from angr.analyses import ReachingDefinitionsAnalysis, CFGFast, CompleteCallingConventionsAnalysis
from angr.code_location import CodeLocation, ExternalCodeLocation
from angr.analyses.reaching_definitions.rd_state import ReachingDefinitionsState
from angr.analyses.reaching_definitions.subject import Subject
from angr.analyses.reaching_definitions.dep_graph import DepGraph
from angr.block import Block
from angr.knowledge_plugins.key_definitions.live_definitions import LiveDefinitions
from angr.knowledge_plugins.key_definitions.atoms import AtomKind, GuardUse, Tmp, Register, MemoryLocation
from angr.knowledge_plugins.key_definitions.constants import ObservationPointType, OP_BEFORE, OP_AFTER
from angr.knowledge_plugins.key_definitions.live_definitions import Definition, SpOffset
from angr.storage.memory_mixins import MultiValuedMemory
from angr.storage.memory_object import SimMemoryObject
from angr.utils.constants import DEFAULT_STATEMENT


class InsnAndNodeObserveTestingUtils:
    @staticmethod
    def assert_for_live_definitions(assertion, live_definition_1, live_definition_2):
        list(
            map(
                lambda attr: {
                    assertion(getattr(live_definition_1, attr)._pages, getattr(live_definition_2, attr)._pages)
                },
                ["registers", "stack", "memory"],
            )
        )
        assertion(getattr(live_definition_1, "tmps"), getattr(live_definition_2, "tmps"))

    @staticmethod
    def filter(observed_results, observation_points):
        # Return only the observed results associated with the observation points,
        # and do not fail if an observation point do not appear in the observed results.
        return list(
            map(lambda key: observed_results[key], filter(lambda key: key in observed_results, observation_points))
        )

    @staticmethod
    def setup(observation_points):
        binary_path = _binary_path("all")
        project = angr.Project(binary_path, load_options={"auto_load_libs": False})
        cfg = project.analyses[CFGFast].prep()()

        main_function = cfg.kb.functions["main"]
        reaching_definitions = project.analyses[ReachingDefinitionsAnalysis].prep()(
            subject=main_function,
            observation_points=observation_points,
        )

        state = ReachingDefinitionsState(
            CodeLocation(main_function.addr, None), project.arch, reaching_definitions.subject
        )

        return (project, main_function, reaching_definitions, state)


def _binary_path(binary_name, arch: str = "x86_64"):
    tests_location = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "..", "..", "..", "..", "binaries", "tests"
    )
    return os.path.join(tests_location, arch, binary_name)


def _result_path(binary_results_name):
    return os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "results", "x86_64", binary_results_name + ".pickle"
    )


class TestReachingDefinitions(TestCase):
    def _run_reaching_definition_analysis_test(self, project, function, result_path, _extract_result):
        tmp_kb = angr.KnowledgeBase(project)
        reaching_definition = project.analyses[ReachingDefinitionsAnalysis].prep(kb=tmp_kb)(
            subject=function,
            observe_all=True,
        )

        result = _extract_result(reaching_definition)

        # Uncomment these to regenerate the reference results... if you dare
        # with open(result_path, "wb") as result_file:
        #    pickle.dump(result, result_file)
        with open(result_path, "rb") as result_file:
            expected_result = pickle.load(result_file)

        self.assertListEqual(result, expected_result)

    @staticmethod
    def _extract_all_definitions_from_storage(storage: MultiValuedMemory):
        all_defs = []
        for page_id, page in storage._pages.items():
            last_mo = None
            for pos, n in enumerate(page.content):
                if n is not None:
                    if type(n) is not set or len(n) == 1:
                        addr = page_id * 4096 + pos
                        if type(n) is set:
                            mo: SimMemoryObject = next(iter(n))
                        else:
                            mo: SimMemoryObject = n
                        if mo is not last_mo:
                            last_mo = mo
                            all_defs.append((addr, list(LiveDefinitions.extract_defs(mo.object))))

        return all_defs

    def test_reaching_definition_analysis_definitions(self):
        def _result_extractor(rda):
            unsorted_result = map(
                lambda x: {
                    "key": x[0],
                    "register_definitions": self._extract_all_definitions_from_storage(x[1].registers),
                    "stack_definitions": self._extract_all_definitions_from_storage(x[1].stack),
                    "memory_definitions": self._extract_all_definitions_from_storage(x[1].memory),
                },
                [(k, v) for k, v in rda.observed_results.items() if k[0] != "stmt"],
            )
            return list(sorted(unsorted_result, key=lambda x: x["key"]))

        binaries_and_results = list(
            map(
                lambda binary: (_binary_path(binary), _result_path(binary + "_definitions")),
                ["all", "fauxware", "loop"],
            )
        )

        for binary, result_path in binaries_and_results:
            project = angr.Project(binary, load_options={"auto_load_libs": False})
            cfg = project.analyses[CFGFast].prep()()
            function = cfg.kb.functions["main"]

            self._run_reaching_definition_analysis_test(project, function, result_path, _result_extractor)

    def test_reaching_definition_analysis_visited_blocks(self):
        def _result_extractor(rda):
            return list(sorted(rda.visited_blocks, key=lambda b: b.addr))

        binaries_and_results = list(
            map(
                lambda binary: (_binary_path(binary), _result_path(binary + "_visited_blocks")),
                ["all", "fauxware", "loop"],
            )
        )

        for binary, result_path in binaries_and_results:
            project = angr.Project(binary, load_options={"auto_load_libs": False})
            cfg = project.analyses[CFGFast].prep()()
            function = cfg.kb.functions["main"]

            self._run_reaching_definition_analysis_test(project, function, result_path, _result_extractor)

    def test_node_observe(self):
        # Create several different observation points
        observation_points = [("node", 0x42, OP_AFTER), ("insn", 0x43, OP_AFTER)]

        _, _, reaching_definition, state = InsnAndNodeObserveTestingUtils.setup(observation_points)

        reaching_definition.node_observe(0x42, state, OP_AFTER)

        results = InsnAndNodeObserveTestingUtils.filter(reaching_definition.observed_results, observation_points)
        expected_results = [state.live_definitions]

        self.assertEqual(results, expected_results)

    def test_insn_observe_an_ailment_statement(self):
        # Create several different observation points
        observation_points = [("node", 0x42, OP_AFTER), ("insn", 0x43, OP_AFTER)]

        _, main_function, reaching_definition, state = InsnAndNodeObserveTestingUtils.setup(observation_points)

        # Here, the statement content does not matter, neither if it is really in the block or else…
        statement = ailment.statement.DirtyStatement(0, None)
        block = main_function._addr_to_block_node[main_function.addr]  # pylint: disable=W0212

        reaching_definition.insn_observe(0x43, statement, block, state, OP_AFTER)

        results = InsnAndNodeObserveTestingUtils.filter(reaching_definition.observed_results, observation_points)
        expected_results = [state.live_definitions]

        self.assertGreater(len(results), 0)
        list(
            map(
                lambda x: InsnAndNodeObserveTestingUtils.assert_for_live_definitions(self.assertEqual, x[0], x[1]),
                zip(results, expected_results),
            )
        )

    def test_insn_observe_before_an_imark_pyvex_statement(self):
        # Create several different observation points
        observation_points = [("node", 0x42, OP_AFTER), ("insn", 0x43, OP_BEFORE)]

        project, main_function, reaching_definition, state = InsnAndNodeObserveTestingUtils.setup(observation_points)

        code_block = main_function._addr_to_block_node[main_function.addr]  # pylint: disable=W0212
        block = Block(addr=0x43, byte_string=code_block.bytestr, project=project)
        statement = block.vex.statements[0]

        reaching_definition.insn_observe(0x43, statement, block, state, OP_BEFORE)

        results = InsnAndNodeObserveTestingUtils.filter(reaching_definition.observed_results, observation_points)
        expected_results = [state.live_definitions]

        self.assertGreater(len(results), 0)
        list(
            map(
                lambda x: InsnAndNodeObserveTestingUtils.assert_for_live_definitions(self.assertEqual, x[0], x[1]),
                zip(results, expected_results),
            )
        )

    def test_insn_observe_after_a_pyvex_statement(self):
        # Create several different observation points
        observation_points = [("node", 0x42, OP_AFTER), ("insn", 0x43, OP_AFTER)]

        project, main_function, reaching_definition, state = InsnAndNodeObserveTestingUtils.setup(observation_points)

        code_block = main_function._addr_to_block_node[main_function.addr]  # pylint: disable=W0212
        block = Block(addr=0x43, byte_string=code_block.bytestr, project=project)
        # When observing OP_AFTER an instruction, the statement has to be the last of a block
        # (or preceding an IMark)
        statement = block.vex.statements[-1]

        reaching_definition.insn_observe(0x43, statement, block, state, OP_AFTER)

        results = InsnAndNodeObserveTestingUtils.filter(reaching_definition.observed_results, observation_points)
        expected_results = [state.live_definitions]

        self.assertGreater(len(results), 0)
        list(
            map(
                lambda x: InsnAndNodeObserveTestingUtils.assert_for_live_definitions(self.assertEqual, x[0], x[1]),
                zip(results, expected_results),
            )
        )

    def test_reaching_definition_analysis_exposes_its_subject(self):
        binary_path = _binary_path("all")
        project = angr.Project(binary_path, load_options={"auto_load_libs": False})
        cfg = project.analyses[CFGFast].prep()()

        main_function = cfg.kb.functions["main"]
        reaching_definitions = project.analyses[ReachingDefinitionsAnalysis].prep()(subject=main_function)

        self.assertEqual(reaching_definitions.subject.__class__ is Subject, True)

    def test_get_the_sp_from_a_reaching_definition(self):
        binary = _binary_path("all")
        project = angr.Project(binary, auto_load_libs=False)
        cfg = project.analyses[CFGFast].prep()()

        tmp_kb = angr.KnowledgeBase(project)
        main_func = cfg.kb.functions["main"]
        rda = project.analyses[ReachingDefinitionsAnalysis].prep(kb=tmp_kb)(subject=main_func, observe_all=True)

        def _is_right_before_main_node(definition):
            bloc, ins_addr, op_type = definition[0]
            return bloc == "node" and ins_addr == main_func.addr and op_type == OP_BEFORE

        reach_definition_at_main = next(filter(_is_right_before_main_node, rda.observed_results.items()))[1]

        sp_value = reach_definition_at_main.get_sp()

        self.assertEqual(sp_value, LiveDefinitions.INITIAL_SP_64BIT)

    def test_dep_graph(self):
        project = angr.Project(_binary_path("true"), auto_load_libs=False)
        cfg = project.analyses[CFGFast].prep()()
        main_func = cfg.functions["main"]

        # build a def-use graph for main() of /bin/true without tmps.
        # check that the only dependency of the first block's
        # guard is the four cc registers
        rda = project.analyses[ReachingDefinitionsAnalysis].prep()(
            subject=main_func, track_tmps=False, track_consts=False, dep_graph=True
        )
        guard_use = list(
            filter(
                lambda def_: type(def_.atom) is GuardUse and def_.codeloc.block_addr == main_func.addr,
                rda.dep_graph._graph.nodes(),
            )
        )[0]
        preds = list(rda.dep_graph._graph.pred[guard_use])
        self.assertEqual(len(preds), 1)
        self.assertIsInstance(preds[0].atom, Register)
        self.assertEqual(
            preds[0].atom.reg_offset,
            project.arch.registers["rdi"][0],
        )

        # build a def-use graph for main() of /bin/true. check that t7 in the first block is only used by the guard
        rda = project.analyses[ReachingDefinitionsAnalysis].prep()(
            subject=main_func, track_tmps=True, dep_graph=DepGraph()
        )
        tmp_7 = list(
            filter(
                lambda def_: type(def_.atom) is Tmp
                and def_.atom.tmp_idx == 7
                and def_.codeloc.block_addr == main_func.addr,
                rda.dep_graph._graph.nodes(),
            )
        )[0]
        self.assertEqual(len(rda.dep_graph._graph.succ[tmp_7]), 1)
        self.assertEqual(type(list(rda.dep_graph._graph.succ[tmp_7])[0].atom), GuardUse)

    def test_dep_graph_stack_variables(self):
        bin_path = _binary_path("fauxware")
        project = angr.Project(bin_path, auto_load_libs=False)
        arch = project.arch
        cfg = project.analyses[CFGFast].prep()()
        main_func = cfg.functions["authenticate"]

        rda: ReachingDefinitionsAnalysis = project.analyses[ReachingDefinitionsAnalysis].prep()(
            subject=main_func, track_tmps=False, track_consts=False, dep_graph=True
        )
        dep_graph = rda.dep_graph
        open_rdi = next(
            iter(
                filter(
                    lambda def_: isinstance(def_.atom, Register)
                    and def_.atom.reg_offset == arch.registers["rdi"][0]
                    and def_.codeloc.ins_addr == 0x4006A2,
                    dep_graph._graph.nodes(),
                )
            )
        )

        # 4006A2     mov  rdi, rax
        preds = list(dep_graph._graph.predecessors(open_rdi))
        self.assertEqual(len(preds), 1)
        rax: Definition = preds[0]
        self.assertIsInstance(rax.atom, Register)
        self.assertEqual(rax.atom.reg_offset, arch.registers["rax"][0])
        self.assertEqual(rax.codeloc.ins_addr, 0x400699)

        # 400699     mov  rax, [rbp+file]
        preds = list(dep_graph._graph.predecessors(rax))
        self.assertEqual(len(preds), 1)
        file_var: Definition = preds[0]
        self.assertIsInstance(file_var.atom, MemoryLocation)
        self.assertIsInstance(file_var.atom.addr, SpOffset)
        self.assertEqual(file_var.atom.addr.offset, -32)
        self.assertEqual(file_var.codeloc.ins_addr, 0x40066C)

        # 40066C     mov  [rbp+file], rdi
        preds = list(dep_graph._graph.predecessors(file_var))
        self.assertEqual(len(preds), 1)
        rdi: Definition = preds[0]
        self.assertIsInstance(rdi.atom, Register)
        self.assertEqual(rdi.atom.reg_offset, arch.registers["rdi"][0])
        self.assertIsInstance(rdi.codeloc, ExternalCodeLocation)

    def test_uses_function_call_arguments(self):
        bin_path = _binary_path("fauxware")
        project = angr.Project(bin_path, auto_load_libs=False)
        arch = project.arch
        cfg = project.analyses[CFGFast].prep()()
        main_func = cfg.functions["main"]

        project.analyses[CompleteCallingConventionsAnalysis].prep()(recover_variables=True)
        rda = project.analyses[ReachingDefinitionsAnalysis].prep()(subject=main_func, track_tmps=False)

        # 4007ae
        # rsi and rdi are all used by authenticate()
        code_location = CodeLocation(0x4007A0, DEFAULT_STATEMENT, ins_addr=0x4007AE)
        uses = rda.all_uses.get_uses_by_location(code_location)
        self.assertEqual(len(uses), 2)
        auth_rdi = next(
            iter(
                filter(
                    lambda def_: isinstance(def_.atom, Register) and def_.atom.reg_offset == arch.registers["rdi"][0],
                    uses,
                )
            )
        )
        auth_rsi = next(
            iter(
                filter(
                    lambda def_: isinstance(def_.atom, Register) and def_.atom.reg_offset == arch.registers["rsi"][0],
                    uses,
                )
            )
        )

        # 4007AB mov     rdi, rax
        self.assertEqual(auth_rdi.codeloc.ins_addr, 0x4007AB)

        # 4007A8 mov     rsi, rdx
        self.assertEqual(auth_rsi.codeloc.ins_addr, 0x4007A8)

    def test_rda_on_a_block_without_cfg(self):
        bin_path = _binary_path("fauxware")
        project = angr.Project(bin_path, auto_load_libs=False)

        block = project.factory.block(project.entry, cross_insn_opt=False)
        _ = project.analyses[ReachingDefinitionsAnalysis].prep()(subject=block, track_tmps=False)  # it should not crash

    def test_partial_register_read(self):
        bin_path = _binary_path("fauxware")
        project = angr.Project(bin_path, auto_load_libs=False)
        cfg = project.analyses[CFGFast].prep()()
        rda = project.analyses[ReachingDefinitionsAnalysis].prep()(subject=cfg.kb.functions["main"], observe_all=True)
        mv = rda.model.observed_results[("insn", 0x400765, OP_BEFORE)].registers.load(
            project.arch.registers["edx"][0],
            size=4,
            endness=project.arch.register_endness,
        )
        assert claripy.is_true(mv.one_value() == claripy.BVV(1, 32))

    def test_conditional_return(self):
        bin_path = _binary_path("check_dap", arch="armel")
        project = angr.Project(bin_path, auto_load_libs=False)
        cfg = project.analyses[CFGFast].prep()(normalize=True)
        rda = project.analyses[ReachingDefinitionsAnalysis].prep()(subject=cfg.kb.functions[0x93E0], observe_all=True)
        sp_0 = rda.model.observed_results[("insn", 0x9410, OP_BEFORE)].registers.load(
            project.arch.sp_offset,
            size=4,
            endness=project.arch.register_endness,
        )
        sp_1 = rda.model.observed_results[("insn", 0x9410, OP_AFTER)].registers.load(
            project.arch.sp_offset,
            size=4,
            endness=project.arch.register_endness,
        )
        assert sp_0 == sp_1

    def test_constants_not_stored_to_live_memory_defs(self):
        # Ensure constants loaded from read-only sections are not stored back to memory definitions. If they are stored,
        # we may accidentally pair them with TOP during state merging.
        project = angr.Project(_binary_path("two_cond_func_call_with_const_arg", "armel"), auto_load_libs=False)
        project.analyses.CFGFast()
        project.analyses.CompleteCallingConventions(recover_variables=True)
        rda = project.analyses.ReachingDefinitions("main", observe_all=True)

        for info in rda.callsites_to("f"):
            (defn,) = info.args_defns[0]
            ld = rda.model.get_observation_by_insn(info.callsite, ObservationPointType.OP_BEFORE)

            # Expect a singular mem predecessor
            preds = rda.dep_graph.find_all_predecessors(defn, kind=AtomKind.MEMORY)
            assert len(preds) == 1

            # Verify not stored
            with self.assertRaises(angr.errors.SimMemoryMissingError):
                atom = preds[0].atom
                ld.memory.load(atom.addr, atom.size)

            # Verify expected constant value
            assert ld.get_concrete_value_from_definition(defn) == 1337


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
__package__ = __package__ or "tests.sim"  # pylint:disable=redefined-builtin

import os
import unittest

import angr

from ..common import bin_location


test_location = os.path.join(bin_location, "tests")

arch_data = {  # (steps, [hit addrs], finished)
    "x86_64": (330, (0x1021C20, 0x1021980, 0x1021BE0, 0x4004B0, 0x400440, 0x400570), True),
    "i386": (
        425,
        (0x90198E0, 0x90195C0, 0x9019630, 0x90198A0, 0x8048370, 0x80482F8, 0x8048440, 0x804846D, 0x8048518),
        True,
    ),
    "ppc": (381, (0x11022F50, 0x11022EB0, 0x10000340, 0x100002E8, 0x1000053C, 0x1000063C), True),
    "ppc64": (372, (0x11047490, 0x100003FC, 0x10000368, 0x10000654, 0x10000770), True),
    "mips": (363, (0x1016F20, 0x400500, 0x400470, 0x400640, 0x400750), True),
    "mips64": (390, (0x12103B828, 0x120000870, 0x1200007E0, 0x120000A80, 0x120000B68), True),
    "armel": (370, (0x10154B8, 0x1108244, 0x83A8, 0x8348, 0x84B0, 0x84E4, 0x85E8), True),
    "aarch64": (370, (0x1020B04, 0x400430, 0x4003B8, 0x400538, 0x400570, 0x40062C), True),
}


# pylint: disable=missing-class-docstring
# pylint: disable=no-self-use
class TestAccuracy(unittest.TestCase):
    def _emulate(self, arch, binary, use_sim_procs, steps, hit_addrs, finished):
        # auto_load_libs can't be disabled as the test takes longer time to execute
        p = angr.Project(
            os.path.join(test_location, arch, binary),
            use_sim_procedures=use_sim_procs,
            rebase_granularity=0x1000000,
            load_debug_info=False,
            auto_load_libs=True,
        )
        state = p.factory.full_init_state(
            args=["./test_arrays"],
            add_options={
                angr.options.STRICT_PAGE_ACCESS,
                angr.options.ENABLE_NX,
                angr.options.ZERO_FILL_UNCONSTRAINED_MEMORY,
                angr.options.USE_SYSTEM_TIMES,
            },
        )

        pg = p.factory.simulation_manager(state, resilience=True)
        pg2 = pg.run(until=lambda lpg: len(lpg.active) != 1)

        is_finished = False
        if len(pg2.active) > 0:
            state = pg2.active[0]
        elif len(pg2.deadended) > 0:
            state = pg2.deadended[0]
            is_finished = True
        elif len(pg2.errored) > 0:
            state = pg2.errored[0].state  # ErroredState object!
        else:
            raise ValueError("The result does not contain a state we can use for this test?")

        assert state.history.depth >= steps

        # this is some wonky control flow that asserts that the items in hit_addrs appear in the state in order.
        trace = state.history.bbl_addrs.hardcopy
        reqs = list(hit_addrs)
        while len(reqs) > 0:
            req = reqs.pop(0)
            while True:
                assert len(trace) > 0
                trace_head = trace.pop(0)
                if trace_head == req:
                    break
                assert trace_head not in reqs

        if finished:
            assert is_finished

    def test_windows(self):
        self._emulate(
            "i386", "test_arrays.exe", True, 41, [], False
        )  # blocked on GetLastError or possibly dynamic loading

    def test_x86_64(self):
        steps, hit_addrs, finished = arch_data["x86_64"]
        self._emulate("x86_64", "test_arrays", False, steps, hit_addrs, finished)

    def test_i386(self):
        steps, hit_addrs, finished = arch_data["i386"]
        self._emulate("i386", "test_arrays", False, steps, hit_addrs, finished)

    def test_ppc(self):
        steps, hit_addrs, finished = arch_data["ppc"]
        self._emulate("ppc", "test_arrays", False, steps, hit_addrs, finished)

    def test_ppc64(self):
        steps, hit_addrs, finished = arch_data["ppc64"]
        self._emulate("ppc64", "test_arrays", False, steps, hit_addrs, finished)

    def test_mips(self):
        steps, hit_addrs, finished = arch_data["mips"]
        self._emulate("mips", "test_arrays", False, steps, hit_addrs, finished)

    def test_mips64(self):
        steps, hit_addrs, finished = arch_data["mips64"]
        self._emulate("mips64", "test_arrays", False, steps, hit_addrs, finished)

    def test_armel(self):
        steps, hit_addrs, finished = arch_data["armel"]
        self._emulate("armel", "test_arrays", False, steps, hit_addrs, finished)

    def test_aarch64(self):
        steps, hit_addrs, finished = arch_data["aarch64"]
        self._emulate("aarch64", "test_arrays", False, steps, hit_addrs, finished)

    def test_locale(self):
        # auto_load_libs can't be disabled as the test takes longer time to execute
        p = angr.Project(os.path.join(test_location, "i386", "isalnum"), use_sim_procedures=False, auto_load_libs=True)
        state = p.factory.full_init_state(args=["./isalnum"], add_options={angr.options.STRICT_PAGE_ACCESS})
        pg = p.factory.simulation_manager(state)
        pg2 = pg.run(
            until=lambda lpg: len(lpg.active) != 1, step_func=lambda lpg: lpg if len(lpg.active) == 1 else lpg.prune()
        )
        assert len(pg2.active) == 0
        assert len(pg2.deadended) == 1
        assert pg2.deadended[0].history.events[-1].type == "terminate"
        assert pg2.deadended[0].history.events[-1].objects["exit_code"].concrete_value == 0


if __name__ == "__main__":
    # emulate('armel', 'test_arrays', False, *arch_data['armel'])
    # import sys; sys.exit()
    unittest.main()



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,no-self-use,line-too-long
__package__ = __package__ or "tests.sim.exec_insn"  # pylint:disable=redefined-builtin

import os
import unittest

import angr

from ...common import bin_location


test_location = os.path.join(bin_location, "tests")


class TestRol(unittest.TestCase):
    def test_rol_x86_64(self):
        binary_path = os.path.join(test_location, "x86_64", "test_rol.exe")

        proj = angr.Project(binary_path, auto_load_libs=False)

        initial_state = proj.factory.blank_state(addr=0x401000)
        r_rax = initial_state.solver.BVS("rax", 64)
        initial_state.regs.rax = r_rax

        pg = proj.factory.simulation_manager(initial_state)
        pg.explore(find=0x401013, avoid=0x401010)
        found_state = pg.found[0]

        result = found_state.solver.eval(r_rax)
        assert result == 0x37B7AB70

    def test_rol_i386(self):
        binary_path = os.path.join(test_location, "i386", "test_rol.exe")

        proj = angr.Project(binary_path, auto_load_libs=False)

        initial_state = proj.factory.blank_state(addr=0x401000)
        r_eax = initial_state.solver.BVS("eax", 32)
        initial_state.regs.eax = r_eax

        pg = proj.factory.simulation_manager(initial_state)
        pg.explore(find=0x401013, avoid=0x401010)
        found_state = pg.found[0]

        result = found_state.solver.eval(r_eax)
        assert result == 0x37B7AB70


if __name__ == "__main__":
    unittest.main()



import logging

import claripy

from ..java import JavaSimProcedure
from ...engines.soot.expressions import SimSootExpr_NewArray
from ...engines.soot.values import SimSootValue_ArrayRef, SimSootValue_StringRef

log = logging.getLogger(name=__name__)


class StringConcat(JavaSimProcedure):
    __provides__ = (("java.lang.String", "concat(java.lang.String)"),)

    def run(self, str_1_ref, str_2_ref):  # pylint: disable=arguments-differ
        log.debug(f"Called SimProcedure java.string.concat with args: {str_1_ref} {str_2_ref}")
        str_1 = self.state.memory.load(str_1_ref)
        str_2 = self.state.memory.load(str_2_ref)
        result = claripy.StrConcat(str_1, str_2)
        return result


class StringEquals(JavaSimProcedure):
    __provides__ = (("java.lang.String", "equals(java.lang.Object)"),)

    def run(self, str_ref_1, str_ref_2):  # pylint: disable=unused-argument
        str_1 = self.state.memory.load(str_ref_1)
        str_2 = self.state.memory.load(str_ref_2)
        return self.state.solver.If(str_1 == str_2, self.state.solver.BVV(1, 32), self.state.solver.BVV(0, 32))


class StringSplit(JavaSimProcedure):
    __provides__ = (("java.lang.String", "split(java.lang.String)"),)

    def generate_symbolic_array(self, state, max_length=1000):
        return self.this_ref

    def run(self, this_ref, separator_ref):
        log.debug(f"Called SimProcedure java.lang.String.split with args: {this_ref}, {separator_ref}")
        self.this_ref = this_ref
        this = self.state.memory.load(this_ref)
        separator = self.state.memory.load(separator_ref)

        if this.concrete and separator.concrete:
            # FIXME: escaping should be fixed in claripy
            separator_value = self.state.solver.eval(separator).replace("\\n", "\n")
            values = self.state.solver.eval(this).split(separator_value)
            str_array = SimSootExpr_NewArray.new_array(self.state, "java.lang.String", claripy.BVV(len(values), 32))

            for idx, value in enumerate(values):
                value_ref = SimSootValue_StringRef.new_string(self.state, claripy.StringV(value))
                elem_ref = SimSootValue_ArrayRef(str_array, idx)
                self.state.memory.store(elem_ref, value_ref)

        else:
            str_array = SimSootExpr_NewArray.new_array(self.state, "java.lang.String", claripy.BVS("array_size", 32))
            str_array.add_default_value_generator(self.generate_symbolic_array)

        return str_array


class StringLength(JavaSimProcedure):
    __provides__ = (("java.lang.String", "length()"),)

    def run(self, this_str):
        log.debug(f"Called SimProcedure java.lang.String.length with args: {this_str}")

        return claripy.StrLen(self.state.memory.load(this_str), 32)


class StringCharAt(JavaSimProcedure):
    __provides__ = (("java.lang.String", "charAt(int)"),)

    def run(self, this_str, index):
        log.debug(f"Called SimProcedure java.lang.String.charAt with args: {this_str} {index}")

        char_str = claripy.StrSubstr(index, claripy.BVV(1, 32), self.state.memory.load(this_str))
        return SimSootValue_StringRef.new_string(self.state, char_str)



# pylint:disable=unnecessary-pass
import logging

from ailment.statement import ConditionalJump, Assignment, Jump
from ailment.expression import ITE

from ....utils.graph import subgraph_between_nodes
from ..utils import remove_labels, to_ail_supergraph
from .optimization_pass import OptimizationPass, OptimizationPassStage


_l = logging.getLogger(__name__)


class ITERegionConverter(OptimizationPass):
    """
    Transform regions of the form `if (c) {x = a} else {x = b}` into `x = c ? a : b`.
    """

    ARCHES = ["X86", "AMD64", "ARMEL", "ARMHF", "ARMCortexM", "MIPS32", "MIPS64"]
    PLATFORMS = ["windows", "linux", "cgc"]
    STAGE = OptimizationPassStage.AFTER_GLOBAL_SIMPLIFICATION
    NAME = "Transform ITE-assignment regions into ternary expression assignments"
    DESCRIPTION = __doc__.strip()

    def __init__(self, func, max_updates=10, **kwargs):
        super().__init__(func, **kwargs)
        self._max_updates = max_updates
        self.analyze()

    def _check(self):
        return True, None

    def _analyze(self, cache=None):
        graph_updated = False
        for _ in range(self._max_updates):
            round_update = False
            ite_assign_regions = self._find_ite_assignment_regions()
            if not ite_assign_regions:
                break

            for region_head, region_tail, true_stmt, false_stmt in ite_assign_regions:
                round_update |= self._convert_region_to_ternary_expr(region_head, region_tail, true_stmt, false_stmt)

            if not round_update:
                break

            graph_updated |= True

        if graph_updated:
            self.out_graph = self._graph

    def _find_ite_assignment_regions(self):
        # find all the if-stmt blocks in a graph with no single successor edges
        super_graph = to_ail_supergraph(remove_labels(self._graph))
        if_stmt_blocks = []
        for node in super_graph.nodes():
            if not node.statements:
                continue

            if isinstance(node.statements[-1], ConditionalJump):
                if_stmt_blocks.append(node)

        # re-find the if-stmts blocks in the original graph
        super_if_ids = {(node.statements[-1].ins_addr, node.statements[-1].idx): node for node in if_stmt_blocks}
        super_to_normal_node = {}
        for node in self._graph.nodes():
            if not node.statements:
                continue

            if isinstance(node.statements[-1], ConditionalJump):
                if_stmt = node.statements[-1]
                if (if_stmt.ins_addr, if_stmt.idx) in super_if_ids:
                    super_node = super_if_ids[(if_stmt.ins_addr, if_stmt.idx)]
                    super_to_normal_node[super_node] = node

        # validate each if-stmt block matches a ternary schema
        ite_candidates = []
        for if_stmt_block in if_stmt_blocks:
            if_stmt = if_stmt_block.statements[-1]
            children = list(super_graph.successors(if_stmt_block))
            if len(children) != 2 or children[0] is children[1]:
                continue

            true_child, false_child = None, None
            for child in children:
                if child.addr == if_stmt.true_target.value:
                    true_child = child
                elif child.addr == if_stmt.false_target.value:
                    false_child = child

            if (
                true_child is None
                or false_child is None
                or true_child not in super_graph
                or false_child not in super_graph
            ):
                continue

            # verify the only statements in the two children are assignments
            true_stmts = [stmt for stmt in true_child.statements if not isinstance(stmt, Jump)]
            false_stmts = [stmt for stmt in false_child.statements if not isinstance(stmt, Jump)]
            if len(true_stmts) != 1 or len(false_stmts) != 1:
                continue

            true_stmt = true_stmts[0]
            false_stmt = false_stmts[0]
            if (
                not isinstance(true_stmt, Assignment)
                or not isinstance(false_stmt, Assignment)
                or not true_stmt.dst.likes(false_stmt.dst)
            ):
                continue

            # must contain a single common predecessor
            if (
                len(list(super_graph.predecessors(true_child))) != 1
                or len(list(super_graph.predecessors(false_child))) != 1
            ):
                continue

            # must contain the same common successor
            true_successors = list(super_graph.successors(true_child))
            if len(true_successors) != 1 or true_successors != list(super_graph.successors(false_child)):
                continue
            common_successor = true_successors[0]

            # lastly, normalize the region we will be editing
            region_head = super_to_normal_node.get(if_stmt_block, None)
            tail_blocks = list(self.blocks_by_addr.get(common_successor.addr, []))
            region_tail = tail_blocks[0] if tail_blocks else None
            if region_head is None or region_tail is None:
                continue

            # we have now found a valid ITE-like expression case
            ite_candidates.append((region_head, region_tail, true_stmt, false_stmt))

        return ite_candidates

    def _convert_region_to_ternary_expr(self, region_head, region_tail, true_stmt, false_stmt):
        if region_head not in self._graph or region_tail not in self._graph:
            return False

        #
        # create a new region_head
        #

        new_region_head = region_head.copy()
        ternary_expr = ITE(
            None,
            region_head.statements[-1].condition,
            true_stmt.src,
            false_stmt.src,
            ins_addr=true_stmt.src.ins_addr,
            vex_block_addr=true_stmt.src.vex_block_addr,
            vex_stmt_idx=true_stmt.src.vex_stmt_idx,
        )
        new_assignment = true_stmt.copy()
        new_assignment.src = ternary_expr
        new_region_head.statements[-1] = new_assignment

        #
        # destroy all the old region blocks
        #

        region_nodes = subgraph_between_nodes(self._graph, region_head, [region_tail])
        for node in region_nodes:
            if node is region_head or node is region_tail:
                continue

            self._remove_block(node)

        #
        # update head and tail
        #

        self._update_block(region_head, new_region_head)
        self._graph.add_edge(new_region_head, region_tail)

        return True



import logging
from typing import Optional, DefaultDict, Dict, Tuple, Set, Any, Union, TYPE_CHECKING, Iterable
from collections import defaultdict

import ailment
import pyvex

from angr.analyses import ForwardAnalysis
from ...block import Block
from ...knowledge_plugins.cfg.cfg_node import CFGNode
from ...codenode import CodeNode
from ...engines.light import SimEngineLight
from ...knowledge_plugins.functions import Function
from ...knowledge_plugins.key_definitions import ReachingDefinitionsModel, LiveDefinitions
from ...knowledge_plugins.key_definitions.constants import OP_BEFORE, OP_AFTER, ObservationPointType
from ...code_location import CodeLocation, ExternalCodeLocation
from ...misc.ux import deprecated
from ..forward_analysis.visitors.graph import NodeType
from ..analysis import Analysis
from .engine_ail import SimEngineRDAIL
from .engine_vex import SimEngineRDVEX
from .rd_state import ReachingDefinitionsState
from .rd_initializer import RDAStateInitializer
from .subject import Subject, SubjectType
from .function_handler import FunctionHandler, FunctionCallRelationships
from .dep_graph import DepGraph

if TYPE_CHECKING:
    from typing import Literal

    ObservationPoint = Tuple[
        Literal["insn", "node", "stmt", "exit"], Union[int, Tuple[int, int], Tuple[int, int, int]], ObservationPointType
    ]

l = logging.getLogger(name=__name__)


class ReachingDefinitionsAnalysis(
    ForwardAnalysis[ReachingDefinitionsState, NodeType, object, object], Analysis
):  # pylint:disable=abstract-method
    """
    ReachingDefinitionsAnalysis is a text-book implementation of a static data-flow analysis that works on either a
    function or a block. It supports both VEX and AIL. By registering observers to observation points, users may use
    this analysis to generate use-def chains, def-use chains, and reaching definitions, and perform other traditional
    data-flow analyses such as liveness analysis.

    * I've always wanted to find a better name for this analysis. Now I gave up and decided to live with this name for
      the foreseeable future (until a better name is proposed by someone else).
    * Aliasing is definitely a problem, and I forgot how aliasing is resolved in this implementation. I'll leave this
      as a post-graduation TODO.
    * Some more documentation and examples would be nice.
    """

    def __init__(
        self,
        subject: Union[Subject, ailment.Block, Block, Function, str] = None,
        func_graph=None,
        max_iterations=3,
        track_tmps=False,
        track_consts=True,
        observation_points: "Iterable[ObservationPoint]" = None,
        init_state: ReachingDefinitionsState = None,
        init_context=None,
        state_initializer: Optional["RDAStateInitializer"] = None,
        cc=None,
        function_handler: "Optional[FunctionHandler]" = None,
        observe_all=False,
        visited_blocks=None,
        dep_graph: Union[DepGraph, bool, None] = True,
        observe_callback=None,
        canonical_size=8,
        stack_pointer_tracker=None,
        use_callee_saved_regs_at_return=True,
        interfunction_level: int = 0,
        track_liveness: bool = True,
        func_addr: Optional[int] = None,
    ):
        """
        :param subject:                         The subject of the analysis: a function, or a single basic block
        :param func_graph:                      Alternative graph for function.graph.
        :param max_iterations:                  The maximum number of iterations before the analysis is terminated.
        :param track_tmps:                      Whether or not temporary variables should be taken into consideration
                                                during the analysis.
        :param iterable observation_points:     A collection of tuples of ("node"|"insn", ins_addr, OP_TYPE) defining
                                                where reaching definitions should be copied and stored. OP_TYPE can be
                                                OP_BEFORE or OP_AFTER.
        :param init_state:                      An optional initialization state. The analysis creates and works on a
                                                copy.
                                                Default to None: the analysis then initialize its own abstract state,
                                                based on the given <Subject>.
        :param init_context:                    If init_state is not given, this is used to initialize the context
                                                field of the initial state's CodeLocation. The only default-supported
                                                type which may go here is a tuple of integers, i.e. a callstack.
                                                Anything else requires a custom FunctionHandler.
        :param cc:                              Calling convention of the function.
        :param function_handler:                The function handler to update the analysis state and results on
                                                function calls.
        :param observe_all:                     Observe every statement, both before and after.
        :param visited_blocks:                  A set of previously visited blocks.
        :param dep_graph:                       An initial dependency graph to add the result of the analysis to. Set it
                                                to None to skip dependency graph generation.
        :param canonical_size:                  The sizes (in bytes) that objects with an UNKNOWN_SIZE are treated as
                                                for operations where sizes are necessary.
        :param dep_graph:                       Set this to True to generate a dependency graph for the subject. It will
                                                be available as `result.dep_graph`.
        :param interfunction_level:             The number of functions we should recurse into. This parameter is only
                                                used if function_handler is not provided.
        :param track_liveness:                  Whether to track liveness information. This can consume
                                                sizeable amounts of RAM on large functions. (e.g. ~15GB for a function
                                                with 4k nodes)

        """

        if isinstance(subject, str):
            subject = self.kb.functions[subject]
        if not isinstance(subject, Subject):
            self._subject = Subject(subject, func_graph, cc)
        else:
            self._subject = subject
        self._graph_visitor = self._subject.visitor

        ForwardAnalysis.__init__(
            self, order_jobs=True, allow_merging=True, allow_widening=False, graph_visitor=self._graph_visitor
        )

        self._track_tmps = track_tmps
        self._track_consts = track_consts
        self._max_iterations = max_iterations
        self._observation_points = observation_points
        self._init_state = init_state
        self._canonical_size = canonical_size
        self._use_callee_saved_regs_at_return = use_callee_saved_regs_at_return
        self._func_addr = func_addr

        if dep_graph is None or dep_graph is False:
            self._dep_graph = None
        elif dep_graph is True:
            self._dep_graph = DepGraph()
        else:
            self._dep_graph = dep_graph

        if function_handler is None:
            self._function_handler = FunctionHandler(interfunction_level).hook(self)
        else:
            if interfunction_level != 0:
                l.warning("RDA(interfunction_level=XXX) doesn't do anything if you provide a function handler")
            self._function_handler = function_handler.hook(self)

        if self._init_state is not None:
            self._init_state = self._init_state.copy()
            self._init_state.analysis = self
            # There should never be an initializer needed then
            self._state_initializer = None
        else:
            self._state_initializer = state_initializer

        self._init_context = init_context

        self._observe_all = observe_all
        self._observe_callback = observe_callback

        # sanity check
        if self._observation_points and any(type(op) is not tuple for op in self._observation_points):
            raise ValueError('"observation_points" must be tuples.')

        self._node_iterations: DefaultDict[int, int] = defaultdict(int)

        self.model: ReachingDefinitionsModel = ReachingDefinitionsModel(
            func_addr=self.subject.content.addr if isinstance(self.subject.content, Function) else None,
            track_liveness=track_liveness,
        )

        bp_as_gpr = False
        the_func = None
        if isinstance(self.subject.content, Function):
            the_func = self.subject.content
        else:
            if self._func_addr is not None:
                try:
                    the_func = self.kb.functions.get_by_addr(self._func_addr)
                except KeyError:
                    pass
        if the_func is not None:
            bp_as_gpr = the_func.info.get("bp_as_gpr", False)

        self._engine_vex = SimEngineRDVEX(
            self.project,
            functions=self.kb.functions,
            function_handler=self._function_handler,
        )
        self._engine_ail = SimEngineRDAIL(
            self.project,
            function_handler=self._function_handler,
            stack_pointer_tracker=stack_pointer_tracker,
            use_callee_saved_regs_at_return=self._use_callee_saved_regs_at_return,
            bp_as_gpr=bp_as_gpr,
        )

        self._visited_blocks: Set[Any] = visited_blocks or set()
        self.function_calls: Dict[CodeLocation, FunctionCallRelationships] = {}

        self._analyze()

    @property
    def observed_results(self) -> Dict[Tuple[str, int, int], LiveDefinitions]:
        return self.model.observed_results

    @property
    def all_definitions(self):
        return self.model.all_definitions

    @all_definitions.setter
    def all_definitions(self, v):
        self.model.all_definitions = v

    @property
    def all_uses(self):
        return self.model.all_uses

    @property
    def one_result(self):
        if not self.observed_results:
            raise ValueError("No result is available.")
        if len(self.observed_results) != 1:
            raise ValueError("More than one results are available.")

        return next(iter(self.observed_results.values()))

    @property
    def dep_graph(self) -> DepGraph:
        if self._dep_graph is None:
            raise ValueError(
                "Cannot access dep_graph if the analysis was not configured to generate one. Try passing "
                "dep_graph=True to the RDA constructor."
            )
        return self._dep_graph

    @property
    def visited_blocks(self):
        return self._visited_blocks

    @deprecated(replacement="get_reaching_definitions_by_insn")
    def get_reaching_definitions(self, ins_addr, op_type):
        return self.get_reaching_definitions_by_insn(ins_addr, op_type)

    def get_reaching_definitions_by_insn(self, ins_addr, op_type):
        key = "insn", ins_addr, op_type
        if key not in self.observed_results:
            raise KeyError(
                "Reaching definitions are not available at observation point %s. "
                "Did you specify that observation point?" % str(key)
            )

        return self.observed_results[key]

    def get_reaching_definitions_by_node(self, node_addr, op_type):
        key = "node", node_addr, op_type
        if key not in self.observed_results:
            raise KeyError(
                "Reaching definitions are not available at observation point %s. "
                "Did you specify that observation point?" % str(key)
            )

        return self.observed_results[key]

    def node_observe(
        self,
        node_addr: int,
        state: ReachingDefinitionsState,
        op_type: ObservationPointType,
        node_idx: Optional[int] = None,
    ) -> None:
        """
        :param node_addr:   Address of the node.
        :param state:       The analysis state.
        :param op_type:     Type of the observation point. Must be one of the following: OP_BEFORE, OP_AFTER.
        :param node_idx:    ID of the node. Used in AIL to differentiate blocks with the same address.
        """

        key = None

        observe = False

        if self._observe_all:
            observe = True
            key: ObservationPoint = (
                ("node", node_addr, op_type) if node_idx is None else ("node", (node_addr, node_idx), op_type)
            )
        elif self._observation_points is not None:
            key: ObservationPoint = (
                ("node", node_addr, op_type) if node_idx is None else ("node", (node_addr, node_idx), op_type)
            )
            if key in self._observation_points:
                observe = True
        elif self._observe_callback is not None:
            observe = self._observe_callback("node", addr=node_addr, state=state, op_type=op_type, node_idx=node_idx)
            if observe:
                key: ObservationPoint = (
                    ("node", node_addr, op_type) if node_idx is None else ("node", (node_addr, node_idx), op_type)
                )

        if observe:
            self.observed_results[key] = state.live_definitions

    def insn_observe(
        self,
        insn_addr: int,
        stmt: Union[ailment.Stmt.Statement, pyvex.stmt.IRStmt],
        block: Union[Block, ailment.Block],
        state: ReachingDefinitionsState,
        op_type: ObservationPointType,
    ) -> None:
        """
        :param insn_addr:   Address of the instruction.
        :param stmt:        The statement.
        :param block:       The current block.
        :param state:       The abstract analysis state.
        :param op_type:     Type of the observation point. Must be one of the following: OP_BEORE, OP_AFTER.
        """

        key = None
        observe = False

        if self._observe_all:
            observe = True
            key: ObservationPoint = ("insn", insn_addr, op_type)
        elif self._observation_points is not None:
            key: ObservationPoint = ("insn", insn_addr, op_type)
            if key in self._observation_points:
                observe = True
        elif self._observe_callback is not None:
            observe = self._observe_callback(
                "insn", addr=insn_addr, stmt=stmt, block=block, state=state, op_type=op_type
            )
            if observe:
                key: ObservationPoint = ("insn", insn_addr, op_type)

        if not observe:
            return

        if isinstance(stmt, pyvex.stmt.IRStmt):
            # it's an angr block
            vex_block = block.vex
            # OP_BEFORE: stmt has to be IMark
            if op_type == OP_BEFORE and type(stmt) is pyvex.stmt.IMark:
                self.observed_results[key] = state.live_definitions.copy()
            # OP_AFTER: stmt has to be last stmt of block or next stmt has to be IMark
            elif op_type == OP_AFTER:
                idx = vex_block.statements.index(stmt)
                if idx == len(vex_block.statements) - 1 or type(vex_block.statements[idx + 1]) is pyvex.IRStmt.IMark:
                    self.observed_results[key] = state.live_definitions.copy()
        elif isinstance(stmt, ailment.Stmt.Statement):
            # it's an AIL block
            self.observed_results[key] = state.live_definitions.copy()

    def stmt_observe(
        self,
        stmt_idx: int,
        stmt: Union[ailment.Stmt.Statement, pyvex.stmt.IRStmt],
        block: Union[Block, ailment.Block],
        state: ReachingDefinitionsState,
        op_type: ObservationPointType,
    ) -> None:
        """

        :param stmt_idx:
        :param stmt:
        :param block:
        :param state:
        :param op_type:
        :return:
        """

        key = None
        observe = False

        block_idx = block.idx if isinstance(block, ailment.Block) else None
        if self._observe_all:
            observe = True
            key: ObservationPoint = ("stmt", (block.addr, block_idx, stmt_idx), op_type)
        elif self._observation_points is not None:
            key: ObservationPoint = ("stmt", (block.addr, block_idx, stmt_idx), op_type)
            if key in self._observation_points:
                observe = True
        elif self._observe_callback is not None:
            observe = self._observe_callback(
                "stmt", stmt_idx=stmt_idx, stmt=stmt, block=block, state=state, op_type=op_type
            )
            if observe:
                key: ObservationPoint = ("stmt", (block.addr, block_idx, stmt_idx), op_type)

        if not observe:
            return

        if isinstance(stmt, pyvex.stmt.IRStmt):
            # it's an angr block
            self.observed_results[key] = state.live_definitions.copy()
        elif isinstance(stmt, ailment.Stmt.Statement):
            # it's an AIL block
            self.observed_results[key] = state.live_definitions.copy()

    def exit_observe(
        self,
        node_addr: int,
        exit_stmt_idx: int,
        block: Union[Block, ailment.Block],
        state: ReachingDefinitionsState,
        node_idx: Optional[int] = None,
    ):
        observe = False
        key = None

        if self._observe_all:
            observe = True
            key = (
                ("exit", (node_addr, exit_stmt_idx), ObservationPointType.OP_AFTER)
                if node_idx is None
                else ("exit", (node_addr, node_idx, exit_stmt_idx), ObservationPointType.OP_AFTER)
            )
        elif self._observation_points is not None:
            key = (
                ("exit", (node_addr, exit_stmt_idx), ObservationPointType.OP_AFTER)
                if node_idx is None
                else ("exit", (node_addr, node_idx, exit_stmt_idx), ObservationPointType.OP_AFTER)
            )
            if key in self._observation_points:
                observe = True
        elif self._observe_callback is not None:
            observe = self._observe_callback(
                "exit",
                node_addr=node_addr,
                exit_stmt_idx=exit_stmt_idx,
                block=block,
                state=state,
            )
            if observe:
                key = (
                    ("exit", (node_addr, exit_stmt_idx), ObservationPointType.OP_AFTER)
                    if node_idx is None
                    else ("exit", (node_addr, node_idx, exit_stmt_idx), ObservationPointType.OP_AFTER)
                )

        if not observe:
            return

        self.observed_results[key] = state.live_definitions.copy()

    @property
    def subject(self):
        return self._subject

    #
    # Main analysis routines
    #

    def _pre_analysis(self):
        pass

    def _initial_abstract_state(self, node) -> ReachingDefinitionsState:
        if self._init_state is not None:
            return self._init_state
        else:
            return ReachingDefinitionsState(
                CodeLocation(node.addr, stmt_idx=0, ins_addr=node.addr, context=self._init_context),
                self.project.arch,
                self.subject,
                track_tmps=self._track_tmps,
                track_consts=self._track_consts,
                analysis=self,
                canonical_size=self._canonical_size,
                initializer=self._state_initializer,
            )

    # pylint: disable=no-self-use,arguments-differ
    def _merge_states(self, _node, *states: ReachingDefinitionsState):
        merged_state, merge_occurred = states[0].merge(*states[1:])
        return merged_state, not merge_occurred

    def _run_on_node(self, node, state: ReachingDefinitionsState):
        """

        :param node:    The current node.
        :param state:   The analysis state.
        :return:        A tuple: (reached fix-point, successor state)
        """

        self._visited_blocks.add(node)

        engine: SimEngineLight

        if isinstance(node, ailment.Block):
            block = node
            block_key = (node.addr, node.idx)
            engine = self._engine_ail
        elif isinstance(node, (Block, CodeNode)):
            block = self.project.factory.block(node.addr, node.size, opt_level=1, cross_insn_opt=False)
            engine = self._engine_vex
            block_key = node.addr
        elif isinstance(node, CFGNode):
            if node.is_simprocedure or node.is_syscall:
                return False, state.copy(discard_tmpdefs=True)
            block = node.block
            engine = self._engine_vex
            block_key = node.addr
        else:
            l.warning("Unsupported node type %s.", node.__class__)
            return False, state.copy(discard_tmpdefs=True)

        state = state.copy(discard_tmpdefs=True)
        self.node_observe(node.addr, state.copy(), OP_BEFORE)

        if self.subject.type == SubjectType.Function:
            node_parents = [
                CodeLocation(pred.addr, 0, block_idx=pred.idx if isinstance(pred, ailment.Block) else None)
                for pred in self._graph_visitor.predecessors(node)
            ]
            if node.addr == self.subject.content.addr:
                node_parents += [ExternalCodeLocation()]
            self.model.at_new_block(
                CodeLocation(block.addr, 0, block_idx=block.idx if isinstance(block, ailment.Block) else None),
                node_parents,
            )

        state = engine.process(
            state,
            block=block,
            fail_fast=self._fail_fast,
            visited_blocks=self._visited_blocks,
            dep_graph=self._dep_graph,
            model=self.model,
        )

        self._node_iterations[block_key] += 1

        self.node_observe(node.addr, state, OP_AFTER)

        # update all definitions and all uses
        self.all_definitions |= state.all_definitions
        for use in [state.stack_uses, state.heap_uses, state.register_uses, state.memory_uses]:
            self.all_uses.merge(use)

        # drop definitions and uses because we will not need them anymore
        state.downsize()

        if self._node_iterations[block_key] < self._max_iterations:
            return True, state
        else:
            return False, state

    def _intra_analysis(self):
        pass

    def _post_analysis(self):
        pass

    def callsites_to(self, target: Union[int, str, Function]) -> Iterable[FunctionCallRelationships]:
        if isinstance(target, (str, int)):
            try:
                func_addr = self.project.kb.functions[target].addr
            except KeyError:
                return
        elif isinstance(target, Function):
            func_addr = target.addr
        else:
            raise TypeError(type(target))

        for info in self.function_calls.values():
            if info.target == func_addr:
                yield info



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,no-self-use,line-too-long
__package__ = __package__ or "tests.sim.exec_func"  # pylint:disable=redefined-builtin

import os
import unittest

import angr

from ...common import bin_location

test_location = os.path.join(bin_location, "tests")


class TestStrFuncs(unittest.TestCase):
    def test_strncpy(self):
        # auto_load_libs can't be disabled as the test cases failed.
        strncpy_amd64 = angr.Project(
            os.path.join(test_location, "x86_64", "strncpy"),
            load_options={"auto_load_libs": True},
            exclude_sim_procedures_list=["strncpy"],
        )
        explorer = strncpy_amd64.factory.simulation_manager()
        explorer.explore(find=[0x4005FF])
        s = explorer.found[0]
        result = s.solver.eval(s.memory.load(s.regs.rax, 16), cast_to=bytes)
        assert result == b"why hello there\0"

    def test_strncpy_size(self):
        # auto_load_libs can't be disabled as the test cases failed.
        strncpy_size_amd64 = angr.Project(
            os.path.join(test_location, "x86_64", "strncpy-size"),
            load_options={"auto_load_libs": True},
            exclude_sim_procedures_list=["strncpy"],
        )
        explorer = strncpy_size_amd64.factory.simulation_manager()
        cfg = strncpy_size_amd64.analyses.CFG(objects=[strncpy_size_amd64.loader.main_object], normalize=True)
        explorer.use_technique(angr.exploration_techniques.LoopSeer(cfg=cfg, bound=50))
        explorer.explore(find=[0x40064C])
        s = explorer.found[0]
        result = s.solver.eval(s.memory.load(s.regs.rax, 40), cast_to=bytes)
        assert result == b"just testing things\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"

    def test_strncpy_verify_null(self):
        # auto_load_libs can't be disabled as the test cases failed.
        strncpy_verify_null_amd64 = angr.Project(
            os.path.join(test_location, "x86_64", "strncpy-verify-null"),
            load_options={"auto_load_libs": True},
            exclude_sim_procedures_list=["strncpy"],
        )
        explorer = strncpy_verify_null_amd64.factory.simulation_manager()
        cfg = strncpy_verify_null_amd64.analyses.CFG(
            objects=[strncpy_verify_null_amd64.loader.main_object], normalize=True
        )
        explorer.use_technique(angr.exploration_techniques.LoopSeer(cfg=cfg, bound=50))
        explorer.explore(find=[0x40064C])
        s = explorer.found[0]
        result = s.solver.eval(s.memory.load(s.regs.rax, 40), cast_to=bytes)
        assert result == b"just testing things\0\0\0\0\0\0\0\0\0\0\0\0\0\0AAAAAA\0"

    def test_strstr_and_strncpy(self):
        # auto_load_libs can't be disabled as the test cases failed.
        strstr_and_strncpy_amd64 = angr.Project(
            os.path.join(test_location, "x86_64", "strstr_and_strncpy"),
            load_options={"auto_load_libs": True},
            exclude_sim_procedures_list=["strstr"],
        )
        explorer = strstr_and_strncpy_amd64.factory.simulation_manager()
        cfg = strstr_and_strncpy_amd64.analyses.CFG(
            objects=[strstr_and_strncpy_amd64.loader.main_object], normalize=True
        )
        explorer.use_technique(angr.exploration_techniques.LoopSeer(cfg=cfg, bound=50))
        explorer.explore(find=[0x400657])
        s = explorer.found[0]
        result = s.solver.eval(s.memory.load(s.regs.rax, 15), cast_to=bytes)
        assert result == b"hi th hi there\0"

    def test_strstr(self):
        # auto_load_libs can't be disabled as the test cases failed.
        strstr_amd64 = angr.Project(
            os.path.join(test_location, "x86_64", "strstr"),
            load_options={"auto_load_libs": True},
            exclude_sim_procedures_list=["strstr"],
        )
        explorer = strstr_amd64.factory.simulation_manager()
        explorer.explore(find=[0x4005FB])
        s = explorer.found[0]
        result = s.solver.eval(s.memory.load(s.regs.rax, 9), cast_to=bytes)
        assert result == b"hi there\0"


if __name__ == "__main__":
    unittest.main()



from ..light import VEXMixin
from ....state_plugins import BP_BEFORE, BP_AFTER, NO_OVERRIDE


class SimInspectMixin(VEXMixin):
    # open question: what should be done about the BP_AFTER breakpoints in cases where the engine uses exceptional
    # control flow?
    def _perform_vex_stmt_Dirty_call(self, func_name, ty, args, func=NO_OVERRIDE):
        self.state._inspect(
            "dirty", when=BP_BEFORE, dirty_name=func_name, dirty_args=args, dirty_handler=func, dirty_result=NO_OVERRIDE
        )
        retval = self.state._inspect_getattr("dirty_result", NO_OVERRIDE)
        func = self.state._inspect_getattr("dirty_handler", func)
        args = self.state._inspect_getattr("dirty_args", args)

        if func is NO_OVERRIDE:
            func = None
        if retval is NO_OVERRIDE:
            retval = super()._perform_vex_stmt_Dirty_call(func_name, ty, args, func=func)

        self.state._inspect("dirty", when=BP_AFTER, dirty_result=retval)
        return self.state._inspect_getattr("dirty_result", retval)

    def _handle_vex_stmt_IMark(self, stmt):
        if self.stmt_idx != 0:
            self.state._inspect("instruction", BP_AFTER)
        super()._handle_vex_stmt_IMark(stmt)
        self.state._inspect("instruction", BP_BEFORE, instruction=stmt.addr + stmt.delta)

    def _handle_vex_expr(self, expr):
        self.state._inspect("expr", BP_BEFORE, expr=expr, expr_result=NO_OVERRIDE)
        expr_result = self.state._inspect_getattr("expr_result", NO_OVERRIDE)
        if expr_result is not NO_OVERRIDE:
            return expr_result
        return super()._handle_vex_expr(expr)

    def _instrument_vex_expr(self, result):
        result = super()._instrument_vex_expr(result)
        self.state._inspect("expr", BP_AFTER, expr_result=result)
        return self.state._inspect_getattr("expr_result", result)

    def _handle_vex_stmt(self, stmt):
        self.state._inspect("statement", BP_BEFORE, statement=self.stmt_idx)
        super()._handle_vex_stmt(stmt)
        self.state._inspect("statement", BP_AFTER)

    def handle_vex_block(self, irsb):
        self.state._inspect("irsb", BP_BEFORE, address=irsb.addr)
        super().handle_vex_block(irsb)
        self.state._inspect("instruction", BP_AFTER)
        self.state._inspect("irsb", BP_AFTER, address=irsb.addr)



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,no-self-use,line-too-long

import unittest

import angr


class TestVariableRegistration(unittest.TestCase):
    def test_registration(self):
        s = angr.SimState(arch="AMD64")

        a1 = s.solver.BVS("a", 64, key=(1,), eternal=True)
        a2 = s.solver.BVS("a", 64, key=(1,), eternal=True)
        assert a1 is a2

        b1 = s.solver.BVS("b", 64, key=(2,), eternal=False)
        s1 = s.copy()
        s2 = s.copy()

        b2 = s1.solver.BVS("b", 64, key=(2,), eternal=False)
        b3 = s2.solver.BVS("b", 64, key=(2,), eternal=False)
        assert b1 is not b2
        assert b2 is not b3
        assert b1 is not b3

        a3 = s1.solver.BVS("a", 64, key=(1,), eternal=True)
        a4 = s2.solver.BVS("a", 64, key=(1,), eternal=True)
        assert a2 is a3
        assert a3 is a4

        assert len(list(s.solver.get_variables(1))) == 1
        assert len(list(s1.solver.get_variables(1))) == 1
        assert len(list(s2.solver.get_variables(1))) == 1

        assert len(list(s.solver.get_variables(2))) == 1
        assert len(list(s1.solver.get_variables(2))) == 2
        assert len(list(s2.solver.get_variables(2))) == 2

        assert list(s.solver.describe_variables(a1)) == [(1,)]
        assert list(s.solver.describe_variables(b1)) == [(2, 1)]
        assert sorted(list(s.solver.describe_variables(a1 + b1))) == [(1,), (2, 1)]


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,no-self-use,line-too-long
__package__ = __package__ or "tests.sim.exec_insn"  # pylint:disable=redefined-builtin

import os
import subprocess
import sys
from unittest import main, skipUnless, TestCase

import angr

from ...common import bin_location


test_location = os.path.join(bin_location, "tests")


class TestSignedDiv(TestCase):
    @skipUnless(sys.platform.startswith("linux"), "linux only")
    def test_signed_div(self):
        test_bin = os.path.join(test_location, "x86_64", "test_signed_div")
        b = angr.Project(test_bin, auto_load_libs=False)

        pg = b.factory.simulation_manager()
        pg.explore()
        out_angr = pg.deadended[0].posix.dumps(1)
        with subprocess.Popen(test_bin, stdout=subprocess.PIPE) as proc:
            stdout_real, _ = proc.communicate()

        assert out_angr == stdout_real


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,no-self-use,line-too-long

import unittest

from angr import SimState


class TestMmap(unittest.TestCase):
    def test_mmap_base_copy(self):
        state = SimState(arch="AMD64", mode="symbolic")

        mmap_base = 0x12345678

        state.heap.mmap_base = mmap_base

        # Sanity check
        assert state.heap.mmap_base == mmap_base

        state_copy = state.copy()

        assert state_copy.heap.mmap_base == mmap_base


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
__package__ = __package__ or "tests.analyses"  # pylint:disable=redefined-builtin

import os
import unittest

import networkx

import angr
from angr.analyses.cdg import TemporaryNode
from angr.utils.graph import compute_dominance_frontier

from ..common import bin_location


test_location = os.path.join(bin_location, "tests")


# pylint: disable=missing-class-docstring
# pylint: disable=no-self-use
class TestCdg(unittest.TestCase):
    def test_graph_0(self):
        # This graph comes from Fig.1 of paper An Efficient Method of Computing Static Single Assignment Form by Ron
        # Cytron, etc.

        # Create a project with a random binary - it will not be used anyways
        p = angr.Project(
            os.path.join(test_location, "x86_64", "datadep_test"),
            load_options={"auto_load_libs": False},
            use_sim_procedures=True,
        )

        # Create the CDG analysis
        cfg = p.analyses.CFGEmulated(no_construct=True)

        # Create our mock control flow graph
        g = networkx.DiGraph()
        edges = [
            ("Entry", 1),
            (1, 2),
            (2, 3),
            (2, 7),
            (3, 4),
            (3, 5),
            (4, 6),
            (5, 6),
            (6, 8),
            (7, 8),
            (8, 9),
            (9, 10),
            (9, 11),
            (11, 9),
            (10, 11),
            (11, 12),
            (12, 2),
            (12, "Exit"),
            ("Entry", "Exit"),
        ]

        for src, dst in edges:
            # Create a TemporaryNode for each node
            n1 = TemporaryNode(src)
            n2 = TemporaryNode(dst)
            g.add_edge(n1, n2)

        # Manually set the CFG
        cfg.model.graph = g
        cfg.model._nodes = {}
        cfg._edge_map = {}
        cfg._loop_back_edges = []
        cfg._overlapped_loop_headers = []

        # Call _construct()
        cdg = p.analyses.CDG(cfg=cfg, no_construct=True)
        cdg._entry = TemporaryNode("Entry")
        cdg._construct()

        standard_result = {
            "Entry": {1, 2, 8, 9, 11, 12},
            1: set(),
            2: {3, 6, 7},
            3: {4, 5},
            4: set(),
            5: set(),
            6: set(),
            7: set(),
            8: set(),
            9: {10},
            10: set(),
            11: {9, 11},
            12: {2, 8, 9, 11, 12},
        }

        for node, cd_nodes in standard_result.items():
            # Each node in set `cd_nodes` is control dependent on `node`
            for n in cd_nodes:
                assert cdg.graph.has_edge(TemporaryNode(node), TemporaryNode(n))
            assert len(cdg.graph.out_edges(TemporaryNode(node))) == len(cd_nodes)

    def test_dominance_frontiers(self):
        # This graph comes from Fig.1 of paper An Efficient Method of Computing Static Single Assignment Form by Ron
        # Cytron, etc.

        # Create our mock control flow graph
        g = networkx.DiGraph()
        g.add_edge("Entry", 1)
        g.add_edge(1, 2)
        g.add_edge(2, 3)
        g.add_edge(2, 7)
        g.add_edge(3, 4)
        g.add_edge(3, 5)
        g.add_edge(4, 6)
        g.add_edge(5, 6)
        g.add_edge(6, 8)
        g.add_edge(7, 8)
        g.add_edge(8, 9)
        g.add_edge(9, 10)
        g.add_edge(9, 11)
        g.add_edge(11, 9)
        g.add_edge(10, 11)
        g.add_edge(11, 12)
        g.add_edge(12, 2)
        g.add_edge(12, "Exit")
        g.add_edge("Entry", "Exit")

        # Create the mock post-dom graph
        postdom = networkx.DiGraph()
        postdom.add_edge("Entry", 1)
        postdom.add_edge(1, 2)
        postdom.add_edge(2, 3)
        postdom.add_edge(3, 4)
        postdom.add_edge(3, 5)
        postdom.add_edge(3, 6)
        postdom.add_edge(2, 7)
        postdom.add_edge(2, 8)
        postdom.add_edge(8, 9)
        postdom.add_edge(9, 10)
        postdom.add_edge(9, 11)
        postdom.add_edge(11, 12)
        postdom.add_edge("Entry", "Exit")

        # Call df_construct()
        df = compute_dominance_frontier(g, postdom)

        standard_df = {
            1: {"Exit"},
            2: {"Exit", 2},
            3: {8},
            4: {6},
            5: {6},
            6: {8},
            7: {8},
            8: {"Exit", 2},
            9: {"Exit", 2, 9},
            10: {11},
            11: {"Exit", 2, 9},
            12: {"Exit", 2},
            "Entry": set(),
            "Exit": set(),
        }
        assert df == standard_df


if __name__ == "__main__":
    unittest.main()



import angr


class b64_decode(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, src, dst, length):
        strncpy = angr.SIM_PROCEDURES["libc"]["strncpy"]

        cpy = self.inline_call(strncpy, dst, src, length)
        self.state.memory.store(dst + 16, self.state.solver.BVV(0, 8))
        return cpy.ret_expr



from typing import Dict, Optional, Set, Union
from sortedcontainers import SortedDict

from ....sim_variable import SimVariable

#
#   Position Mapping Classes
#


class PositionMappingElement:
    __slots__ = ("start", "length", "obj")

    def __init__(self, start, length, obj):
        self.start: int = start
        self.length: int = length
        self.obj = obj

    def __contains__(self, offset):
        return self.start <= offset < self.start + self.length

    def __repr__(self):
        return "<%d-%d: %s>" % (self.start, self.start + self.length, self.obj)


class PositionMapping:
    __slots__ = ("_posmap",)

    DUPLICATION_CHECK = True

    def __init__(self):
        self._posmap: Union[SortedDict, Dict[int, PositionMappingElement]] = SortedDict()

    def items(self):
        return self._posmap.items()

    #
    # Public methods
    #

    def add_mapping(self, start_pos, length, obj):
        # duplication check
        if self.DUPLICATION_CHECK:
            try:
                pre = next(self._posmap.irange(maximum=start_pos, reverse=True))
                if start_pos in self._posmap[pre]:
                    raise ValueError("New mapping is overlapping with an existing element.")
            except StopIteration:
                pass

        self._posmap[start_pos] = PositionMappingElement(start_pos, length, obj)

    def get_node(self, pos: int):
        element = self.get_element(pos)
        if element is None:
            return None
        return element.obj

    def get_element(self, pos: int) -> Optional[PositionMappingElement]:
        try:
            pre = next(self._posmap.irange(maximum=pos, reverse=True))
        except StopIteration:
            return None

        element = self._posmap[pre]
        if pos in element:
            return element
        return None


class InstructionMappingElement:
    __slots__ = ("ins_addr", "posmap_pos")

    def __init__(self, ins_addr, posmap_pos):
        self.ins_addr: int = ins_addr
        self.posmap_pos: int = posmap_pos

    def __contains__(self, offset: int):
        return self.ins_addr == offset

    def __repr__(self):
        return "<%d: %d>" % (self.ins_addr, self.posmap_pos)


class InstructionMapping:
    __slots__ = ("_insmap",)

    def __init__(self):
        self._insmap: Union[SortedDict, Dict[int, InstructionMappingElement]] = SortedDict()

    def items(self):
        return self._insmap.items()

    def add_mapping(self, ins_addr, posmap_pos):
        if ins_addr in self._insmap:
            if posmap_pos <= self._insmap[ins_addr].posmap_pos:
                self._insmap[ins_addr] = InstructionMappingElement(ins_addr, posmap_pos)
        else:
            self._insmap[ins_addr] = InstructionMappingElement(ins_addr, posmap_pos)

    def get_nearest_pos(self, ins_addr: int) -> Optional[int]:
        try:
            pre_max = next(self._insmap.irange(maximum=ins_addr, reverse=True))
            pre_min = next(self._insmap.irange(minimum=ins_addr, reverse=True))
        except StopIteration:
            return None

        e1: InstructionMappingElement = self._insmap[pre_max]
        e2: InstructionMappingElement = self._insmap[pre_min]

        if abs(ins_addr - e1.ins_addr) <= abs(ins_addr - e2.ins_addr):
            return e1.posmap_pos
        else:
            return e2.posmap_pos


class BaseStructuredCodeGenerator:
    def __init__(self, flavor=None):
        self.flavor = flavor
        self.text = None
        self.map_pos_to_node = None
        self.map_pos_to_addr = None
        self.map_addr_to_pos = None
        self.map_ast_to_pos: Optional[Dict[SimVariable, Set[PositionMappingElement]]] = None

    def reapply_options(self, options):
        pass

    def regenerate_text(self) -> None:
        pass

    def reload_variable_types(self) -> None:
        pass
